from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(
    trange=["1982-12-06", "1982-12-07"],
    instrument="mag",
    prefix="",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    This function loads data from the DE2 mission.

    Data is available for three years: 1981, 1982, and 1983.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Default is ["1982-12-06", "1982-12-07"].
        Data is available only for years 1981, 1982, 1983.
    instrument : str, optional
        Valid options are: 'mag', 'nacs', 'rpa', 'idm', 'wats', 'vefi', 'lang'.
        Default is 'mag'.
        Each instrument has only one datatype available, so the datatype is not a keyword option.
    prefix : str, optional
        The tplot variable names will be given this prefix.
        Default is no prefix is added.
    suffix : str, optional
        The tplot variable names will be given this suffix.
        Default is no suffix is added.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data"
        will be loaded into tplot.
        Default is only loads in data with a "VAR_TYPE" attribute of "data".
    varformat : str, optional
        The file variable formats to load into tplot.
        Wildcard character "*" is accepted.
        Default is all variables are loaded in.
    varnames : list of str, optional
        List of variable names to load.
        Default is all data variables are loaded.
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables.
        Default is False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables.
        Default is False.
    no_update : bool, optional
        If set, only load data from your local cache.
        Default is False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword.
        Default is False.
    force_download : bool, optional
        Download file even if local version is more recent than server version.
        Default is False.

    Returns
    -------
    list
        List of tplot variables created.
        If notplot is set to True, returns a list of dictionaries containing the data.
        If downloadonly is set to True, returns a list of the downloaded files.


    Examples
    --------
    >>> import pyspedas
    >>> vars = pyspedas.projects.de2.load(instrument='mag', trange=['1983-02-10', '1983-02-11'], datatype='62ms', prefix = 'de2_')
    >>> print(vars)
    ['de2_ex', 'de2_ey', 'de2_bx', 'de2_by', 'de2_bz', 'de2_bxm', 'de2_bym', 'de2_bzm',
    'de2_glat', 'de2_glon', 'de2_ilat', 'de2_mlt', 'de2_alt']

    >>> import pyspedas
    >>> vars = pyspedas.projects.de2.lang(trange=['1983-02-10', '1983-02-11'], prefix = 'de2_')
    >>> print(vars)
    ['de2_OrbitNumber', 'de2_electronTemp', 'de2_plasmaDensity', 'de2_satPotential',
    'de2_alt', 'de2_glat', 'de2_glon', 'de2_lst', 'de2_lmt', 'de2_L', 'de2_ilat', 'de2_sza']
    """

    mastercdf = None
    addmaster = False
    masterpath = "https://cdaweb.gsfc.nasa.gov/pub/software/cdawlib/0MASTERS/"
    local_master_dir = CONFIG["local_data_dir"] + "de2_masters/"
    tvars = []

    if instrument == "mag":
        # https://spdf.gsfc.nasa.gov/pub/data/de/de2/magnetic_electric_fields_vefi_magb/62ms_vefimagb_cdaweb/
        datatype = "62ms"  # this is the only available datatype
        pathformat = (
            "magnetic_electric_fields_vefi_magb/"
            + datatype
            + "_vefimagb_cdaweb/%Y/de2_"
            + datatype
            + "_vefimagb_%Y%m%d_v??.cdf"
        )
    elif instrument == "nacs":
        # https://spdf.gsfc.nasa.gov/pub/data/de/de2/neutral_gas_nacs/neutral1s_nacs_cdaweb/
        datatype = "neutral1s"  # this is the only available datatype
        pathformat = (
            "neutral_gas_nacs/"
            + datatype
            + "_"
            + instrument
            + "_cdaweb/%Y/de2_"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "rpa":
        # https://spdf.gsfc.nasa.gov/pub/data/de/de2/plasma_rpa/ion2s_cdaweb/
        datatype = "ion2s"  # this is the only available datatype
        pathformat = (
            "plasma_rpa/"
            + datatype
            + "_cdaweb/%Y/de2_"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "fpi":
        # https://spdf.gsfc.nasa.gov/pub/data/de/de2/neutral_gas_fpi/de2_neutral8s_fpi/
        datatype = "8s"  # this is the only available datatype
        pathformat = (
            "neutral_gas_fpi/de2_neutral8s_fpi/%Y/de2_neutral"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "idm":
        # https://spdf.gsfc.nasa.gov/pub/data/de/de2/plasma_idm/vion250ms_cdaweb/
        datatype = "250ms"  # this is the only available datatype
        masterfile = "de2_vion250ms_idm_00000000_v01.cdf"
        addmaster = True
        pathformat = (
            "plasma_idm/vion250ms_cdaweb/%Y/de2_vion"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "wats":
        # https://spdf.gsfc.nasa.gov/pub/data/de/de2/neutral_gas_wats/wind2s_wats_cdaweb/
        datatype = "2s"  # this is the only available datatype
        pathformat = (
            "neutral_gas_wats/wind2s_wats_cdaweb/%Y/de2_wind"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "vefi":
        # https://spdf.gsfc.nasa.gov/pub/data/de/de2/electric_fields_vefi/dca500ms_vefi_cdaweb/
        datatype = "ac500ms"  # this is the only available datatype
        pathformat = (
            "electric_fields_vefi/"
            + datatype
            + "_vefi_cdaweb/%Y/de2_"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "lang":
        # https://spdf.gsfc.nasa.gov/pub/data/de/de2/plasma_lang/plasma500ms_lang_cdaweb/
        datatype = "500ms"  # this is the only available datatype
        pathformat = (
            "plasma_lang/plasma500ms_lang_cdaweb/%Y/de2_plasma"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )

    if addmaster:
        mastercdf = download(
            remote_file=masterfile,
            remote_path=masterpath,
            local_path=local_master_dir,
            no_download=no_update,
            force_download=force_download,
        )
    else:
        mastercdf = [None]
    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(
        remote_file=remote_names,
        remote_path=CONFIG["remote_data_dir"],
        local_path=CONFIG["local_data_dir"],
        no_download=no_update,
        force_download=force_download,
    )
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    if len(out_files) > 0:
        tvars = cdf_to_tplot(
            out_files,
            mastercdf=mastercdf[0],
            prefix=prefix,
            suffix=suffix,
            get_support_data=get_support_data,
            varformat=varformat,
            varnames=varnames,
            notplot=notplot,
        )

        if notplot:
            return tvars

        if len(tvars) > 0 and time_clip:
            tclip(tvars, trange[0], trange[1], suffix="")

    return tvars



import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import pyspedas
from pyspedas.tplot_tools import get_data
import pyspedas
from pyspedas.tplot_tools import time_double
from pyspedas.projects.mms.fpi_tools.mms_get_fpi_dist import mms_get_fpi_dist
from pyspedas.projects.mms.fpi_tools.mms_pad_fpi import mms_pad_fpi


def mms_fpi_ang_ang(
    time,
    species="i",
    probe="1",
    data_rate="fast",
    fgm_data_rate="srvy",
    level="l2",
    fgm_level="l2",
    center_measurement=False,
    energy_range=[10, 30000],
    xsize=8,
    ysize=8,
    cmap="spedas",
    zrange=[None, None],
    nocontours=False,
    save_png=None,
    save_jpeg=None,
    save_eps=None,
    save_svg=None,
    dpi=300,
    display=True,
):
    """
    Creates various plots directly from the FPI distribution functions, including:
         - angle-angle (azimuth vs zenith)
         - angle-energy (azimuth and zenith vs energy)

    Input
    ------
        time: str
            Time to generate the angle-angle plots for

    Parameters
    -----------
        species: str
            'i' for ions, 'e' for electrons

        probe: str
            MMS spacecraft probe # (1-4)

        data_rate: str
            Data rate ('brst' or 'fast')

        level: str
            Data level

        center_measurement: bool
            Flag to center the FPI measurements

        energy_range: list
            Range of energies to include in the figures (default: 10-30000 eV)

        xsize: float
            X-size of the output figure

        ysize: float
            Y-size of the output figure

        cmap: str
            Color map to use (default: 'spedas')

        zrange: list
            Range of the color bar

        display: bool
            Flag to display the figures (with plt.show()); default: True

    """

    trange = [time_double(time) - 300, time_double(time) + 300]

    fpi_vars = pyspedas.projects.mms.fpi(
        trange=trange,
        datatype=["d" + species + "s-moms", "d" + species + "s-dist"],
        probe=probe,
        data_rate=data_rate,
        level=level,
        center_measurement=center_measurement,
    )

    fgm_vars = pyspedas.projects.mms.fgm(
        trange=trange, probe=probe, data_rate=fgm_data_rate, level=fgm_level
    )
    dist = get_data("mms" + probe + "_d" + species + "s_dist_" + data_rate)

    closest_idx = np.searchsorted(dist.times, time_double(time), side="left")

    data = dist.y[closest_idx, :, :, :]

    if len(dist.v1.shape) > 1:
        phi = dist.v1[closest_idx, :]
    else:
        phi = dist.v1

    theta = dist.v2
    energies = dist.v3[closest_idx, :]
    energy_axis = [np.nanmin(energies), np.nanmax(energies)]

    idx_of_ens = np.argwhere(
        (energies >= energy_range[0]) & (energies <= energy_range[1])
    ).flatten()
    energies = energies[idx_of_ens]
    data_at_ens = data[:, :, idx_of_ens]
    data_summed = np.nansum(data_at_ens, axis=2)

    dists = mms_get_fpi_dist(
        "mms" + probe + "_d" + species + "s_dist_" + data_rate, single_time=time
    )

    # theta is stored as co-latitude
    theta_colat = dists[0]["theta"][0, 0, :]

    # convert to latitude
    theta_flow_direction = 90.0 - theta_colat
    phi = dists[0]["phi"][0, :, 0]

    spec_options = {}

    pa_dist = mms_pad_fpi(
        dists,
        time=dist.times[closest_idx],
        mag_data="mms"
        + probe
        + "_fgm_b_gse_"
        + fgm_data_rate
        + "_"
        + fgm_level
        + "_bvec",
    )

    # the first figure: azimuth vs. zenith
    fig, axes = plt.subplots()
    fig.set_size_inches(xsize, ysize)

    # add support for the 'spedas' color bar
    if cmap == "spedas":
        _colors = pyspedas.tplot_tools.spedas_colorbar
        spd_map = [
            (np.array([r, g, b])).astype(np.float64) / 256
            for r, g, b in zip(_colors.r, _colors.g, _colors.b)
        ]
        spec_options["cmap"] = LinearSegmentedColormap.from_list("spedas", spd_map)
    else:
        spec_options["cmap"] = cmap

    # phi can be out of order, so we need to sort prior to sending to pcolormesh
    phi_sorted_idx = np.argsort(phi)
    phi_sorted = phi[phi_sorted_idx]
    data_phi_sorted = data_summed[phi_sorted_idx, :]

    spec_options["norm"] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    im = axes.pcolormesh(
        phi_sorted, theta_flow_direction, data_phi_sorted.T, **spec_options
    )

    axes.set_xlabel("Az flow angle (deg)")
    axes.set_ylabel("Zenith flow angle (deg)")
    fig.subplots_adjust(left=0.14, right=0.85)
    box = axes.get_position()
    pad, width = 0.02, 0.02
    cax = fig.add_axes([box.xmax + pad, box.ymin, width, box.height])
    colorbar = fig.colorbar(im, cax=cax)
    colorbar.set_label("f ($s^3$/$cm^6$)")

    if not nocontours:
        num_levels = 16
        contour_levels = np.array(
            180 * np.arange(num_levels + 1) / num_levels, dtype=int
        )
        contours = axes.contour(
            pa_dist["wpol"],
            pa_dist["waz"],
            pa_dist["pa_azpol"],
            contour_levels,
            linewidths=0.5,
        )
        axes.clabel(contours, contours.levels, inline=True, fontsize=10)

    if save_png is not None and save_png != "":
        if len(save_png) > 4 and save_png[-4:] == ".png":
            save_png = save_png[:-4]
        plt.savefig(save_png + "_azimuth_vs_zenith.png", dpi=dpi)

    if save_jpeg is not None and save_jpeg != "":
        if len(save_jpeg) > 5 and save_jpeg[-5:] == ".jpeg":
            save_jpeg = save_jpeg[:-5]
        plt.savefig(save_jpeg + "_azimuth_vs_zenith.jpeg", dpi=dpi)

    if save_eps is not None and save_eps != "":
        if len(save_eps) > 4 and save_eps[-4:] == ".eps":
            save_eps = save_eps[:-4]
        plt.savefig(save_eps + "_azimuth_vs_zenith.eps", dpi=dpi)

    if save_svg is not None and save_svg != "":
        if len(save_svg) > 4 and save_svg[-4:] == ".svg":
            save_svg = save_svg[:-4]
        plt.savefig(save_svg + "_azimuth_vs_zenith.svg", dpi=dpi)

    # Zenith vs. energy
    fig2, axes2 = plt.subplots()
    fig2.set_size_inches(xsize, ysize)

    theta_en = np.nansum(data_at_ens, axis=0)

    axes2.set_xscale("log")
    axes2.set_xlabel("Energy (eV)")
    axes2.set_ylabel("Zenith flow angle (deg)")
    fig2.subplots_adjust(left=0.14, right=0.85)

    zrange_is_None = zrange[0] is None

    if zrange_is_None:
        zrange[0] = np.nanmin(theta_en[theta_en > 0])

    spec_options["norm"] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    im2 = axes2.pcolormesh(energies, theta_flow_direction, theta_en, **spec_options)

    cax2 = fig2.add_axes([box.xmax + pad, box.ymin, width, box.height])
    colorbar2 = fig2.colorbar(im2, cax=cax2)
    colorbar2.set_label("f ($s^3$/$cm^6$)")

    if save_png is not None and save_png != "":
        if len(save_png) > 4 and save_png[-4:] == ".png":
            save_png = save_png[:-4]
        plt.savefig(save_png + "_zenith_vs_energy.png", dpi=dpi)

    if save_jpeg is not None and save_jpeg != "":
        if len(save_jpeg) > 5 and save_jpeg[-5:] == ".jpeg":
            save_jpeg = save_jpeg[:-5]
        plt.savefig(save_jpeg + "_zenith_vs_energy.jpeg", dpi=dpi)

    if save_eps is not None and save_eps != "":
        if len(save_eps) > 4 and save_eps[-4:] == ".eps":
            save_eps = save_eps[:-4]
        plt.savefig(save_eps + "_zenith_vs_energy.eps", dpi=dpi)

    if save_svg is not None and save_svg != "":
        if len(save_svg) > 4 and save_svg[-4:] == ".svg":
            save_svg = save_svg[:-4]
        plt.savefig(save_svg + "_zenith_vs_energy.svg", dpi=dpi)

    # Azimuth vs. energy
    fig3, axes3 = plt.subplots()
    fig3.set_size_inches(xsize, ysize)

    phi_en = np.nansum(data_at_ens[phi_sorted_idx, :, :], axis=1)

    fig3.subplots_adjust(left=0.14, right=0.85)
    axes3.set_xscale("log")
    axes3.set_xlabel("Energy (eV)")
    axes3.set_ylabel("Azimuth flow angle (deg)")

    if zrange_is_None:
        zrange[0] = np.nanmin(phi_en[phi_en > 0])

    spec_options["norm"] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    im3 = axes3.pcolormesh(energies, phi_sorted, phi_en, **spec_options)

    cax3 = fig3.add_axes([box.xmax + pad, box.ymin, width, box.height])
    colorbar3 = fig3.colorbar(im3, cax=cax3)
    colorbar3.set_label("f ($s^3$/$cm^6$)")

    if save_png is not None and save_png != "":
        if len(save_png) > 4 and save_png[-4:] == ".png":
            save_png = save_png[:-4]
        plt.savefig(save_png + "_azimuth_vs_energy.png", dpi=dpi)

    if save_jpeg is not None and save_jpeg != "":
        if len(save_jpeg) > 5 and save_jpeg[-5:] == ".jpeg":
            save_jpeg = save_jpeg[:-5]
        plt.savefig(save_jpeg + "_azimuth_vs_energy.jpeg", dpi=dpi)

    if save_eps is not None and save_eps != "":
        if len(save_eps) > 4 and save_eps[-4:] == ".eps":
            save_eps = save_eps[:-4]
        plt.savefig(save_eps + "_azimuth_vs_energy.eps", dpi=dpi)

    if save_svg is not None and save_svg != "":
        if len(save_svg) > 4 and save_svg[-4:] == ".svg":
            save_svg = save_svg[:-4]
        plt.savefig(save_svg + "_azimuth_vs_energy.svg", dpi=dpi)

    # PA vs. energy
    fig4, axes4 = plt.subplots()
    fig4.set_size_inches(xsize, ysize)

    fig4.subplots_adjust(left=0.14, right=0.85)
    axes4.set_yscale("log")
    axes4.set_xlabel("Pitch angle (deg)")
    axes4.set_ylabel("Energy (eV)")

    spec_options["norm"] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    im4 = axes4.pcolormesh(
        pa_dist["pa"], pa_dist["egy"], pa_dist["data"], **spec_options
    )

    cax4 = fig4.add_axes([box.xmax + pad, box.ymin, width, box.height])
    colorbar4 = fig4.colorbar(im4, cax=cax4)
    colorbar4.set_label("f ($s^3$/$cm^6$)")

    if save_png is not None and save_png != "":
        if len(save_png) > 4 and save_png[-4:] == ".png":
            save_png = save_png[:-4]
        plt.savefig(save_png + "_pa_vs_energy.png", dpi=dpi)

    if save_jpeg is not None and save_jpeg != "":
        if len(save_jpeg) > 5 and save_jpeg[-5:] == ".jpeg":
            save_jpeg = save_jpeg[:-5]
        plt.savefig(save_jpeg + "_pa_vs_energy.jpeg", dpi=dpi)

    if save_eps is not None and save_eps != "":
        if len(save_eps) > 4 and save_eps[-4:] == ".eps":
            save_eps = save_eps[:-4]
        plt.savefig(save_eps + "_pa_vs_energy.eps", dpi=dpi)

    if save_svg is not None and save_svg != "":
        if len(save_svg) > 4 and save_svg[-4:] == ".svg":
            save_svg = save_svg[:-4]
        plt.savefig(save_svg + "_pa_vs_energy.svg", dpi=dpi)

    if display:
        plt.show()



from functools import update_wrapper
from .load import load
from pyspedas.utilities.datasets import find_datasets
from pyspedas.utilities.pyspedas_functools import better_partial


# Define partial wrappers for other load routines, fixing the instrument parameter
# 'better_partial' works better with PyCharm autocompletion than functools.partial
# update_wrapper() is necessary for help() to show info for the wrapped function,
# rather than the partial() object.

mag = better_partial(load, instrument="mag")
update_wrapper(mag, load)
fc = better_partial(load, instrument="fc")
update_wrapper(fc, load)
orb = better_partial(load, instrument="orb")
update_wrapper(orb, load)
att = better_partial(load, instrument="att")
update_wrapper(att, load)
all = better_partial(load, instrument="all")
update_wrapper(all, load)

datasets = better_partial(find_datasets, mission="DSCOVR", label=True)
update_wrapper(datasets, find_datasets)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import sys
sys.path.insert(0, os.path.abspath('../..'))
import tomli

import pyspedas
import pyspedas.cotrans_tools.cotrans

# -- Project information -----------------------------------------------------

try:
    with open("../../pyproject.toml", "rb") as f:
        toml_dict = tomli.load(f)

    pyproj_version = toml_dict['project']['version']
except (KeyError, FileNotFoundError):
    pyproj_version = 'unknown'

print(f"Pyproj_version: {pyproj_version}")
project = 'PySPEDAS'
copyright = '2018-2025, Regents of the University of California, unless otherwise indicated'
author = 'The PySPEDAS Community'

# The full version, including alpha/beta/rc tags
release = pyproj_version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.napoleon', 'sphinx.ext.autodoc', 'sphinx.ext.viewcode', 'sphinx.ext.intersphinx']

# Napoleon settings
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = True
napoleon_use_param = False
napoleon_use_rtype = True
napoleon_preprocess_types = True
napoleon_attr_annotations = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = ['css/custom.css']

autodoc_mock_imports = ['sip', 'PyQt5', 'PyQt5.QtGui', 'PyQt5.QtCore', 'PyQt5.QtWidgets']

# Intersphinx generates automatic links to the documentation of objects
# in other packages. When mappings are removed or added, please update
# the section in docs/doc_guide.rst on references to other packages.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "sphinx_automodapi": (
        "https://sphinx-automodapi.readthedocs.io/en/latest/",
        None,
    ),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}


# Load and show tplot variable created using load_fdsn

from pyspedas.mth5.load_fdsn import load_fdsn
import pyspedas
import numpy as np

load_fdsn(network="4P", station="ALW48", trange=['2015-06-22', '2015-06-24'])
pyspedas.tplot('fdsn_4P_ALW48')

# Get data
time, mag = pyspedas.get_data('fdsn_4P_ALW48')

# Re-normalize magnetometer data removing the baseline
mag -= np.mean(mag, axis=0)

pyspedas.store_data('fdsn_4P_ALW48_norm', data={'x': time, 'y':mag})

# Plot
pyspedas.tplot('fdsn_4P_ALW48_norm')



import numpy as np
import math
import logging
from typing import Dict
from .spinmodel_segment import SpinmodelSegment
from pyspedas.tplot_tools import get_data, store_data
from pyspedas.tplot_tools import data_exists
from pyspedas.tplot_tools import time_string


def get_sm_data(probe: str,
                valname: str,
                correction_level: int,
                suffix: str = '') -> tuple[np.ndarray, np.ndarray] | None:
    """ Return the times and values for a spin model tplot variable

    Args:
        probe: Probe name, one of 'a','b','c','d','e'
        valname: The data quantity to be returned
        correction_level: 0 for no corrections, 1 for waveform corrections, 2 for spin fit corrections
        suffix: Suffix to be added to variable names

    Returns:
        A tuple containing the timestamps and data values as np.ndarray objects
    """
    if correction_level == 0:
        infix = '_'
    else:
        infix = '_ecl_'
    tvar_name = 'th' + probe + '_spin' + infix + valname + suffix
    if data_exists(tvar_name):
        res = get_data(tvar_name)
        return res.times, res.y
    else:
        return None


class SpinmodelInterpTResult:
    """ An object to return the results of interpolating a spin model.

    Attributes:
        spinphase (ndarray(dtype=float)): The spin phase (in degrees) at each input time
        spincount (ndarray(dtype=int)): The count of complete spins from the start of the model, at each input time
        spinper (ndarray(dtype=float)): The spin period (in seconds) at each input time
        t_last (ndarray(dtype=float)): Time (seconds since epoch) of last sun sensor crossing before the input times
        eclipse_delta_phi (ndarray(dtype=float)): The offset in degrees between the modeled spin phase and the onboard
            spin phase during an eclipse
        segflags (ndarray(dtype=int)): A set of bit flags denoting the eclipse and correction status at each input time
        idx (ndarray(dtype=int)): The index of the spin model segment containing each input time
        dt (ndarray(dtype=float)): The delta-t in seconds between the input times and start or end time of the
            containing segment

    """

    __slots__ = 'spinphase', 'spincount', 'spinper', 't_last', 'eclipse_delta_phi', 'segflags', 'idx', 'dt'

    def __init__(self, spinphase: np.ndarray,
                 spincount: np.ndarray,
                 spinper: np.ndarray,
                 t_last: np.ndarray,
                 eclipse_delta_phi: np.ndarray,
                 segflags: np.ndarray,
                 idx: np.ndarray,
                 dt: np.ndarray):
        self.spinphase = spinphase
        self.spincount = spincount
        self.spinper = spinper
        self.t_last = t_last
        self.eclipse_delta_phi = eclipse_delta_phi
        self.segflags = segflags
        self.idx = idx
        self.dt = dt


class Spinmodel:
    """ An object describing a spin model for a given probe, time interval, and correction level
    A spinmodel is created from a set of tplot variables loaded from the THEMIS STATE CDFs. The time interval is
    covered by a list of SpinmodelSegment objects, representing a piecewise quadratic fit of phi (spin plane angle)
    versus time.  The segment data from the CDF is loaded into the seg_list during initial creation of the model, with
    some on-the-fly adjustments to account for the brief gaps that occur at UTC date boundaries.  The final segment
    list is converted to a series of ndarrays, stored in the seg_* attributes, to support efficient interpolation to
    input times.

    Attributes:
        seg_list (List of SpinmodelSegment): A list of spin model segment objects comprising the model
        lastseg (SpinmodelSegment): The most recently processed SpinmodelSegment
        seg_times (ndarray): Array of t1 values from seg_list
        seg_t2 (ndarray): Array of t2 values from seg_list
        seg_c1 (ndarray): Array of c1 values from seg_list
        seg_c2 (ndarray): Array of c2 values from seg_list
        seg_b (ndarray): Array of b values from seg_list
        seg_c (ndarray): Array of c valuse from seg_list
        seg_npts (ndarray): Array of npts values from seg_list
        seg_maxgap (ndarray): Array of maxgap values from seg_list
        seg_phaserr (ndarray): Array of phaserr values from seg_list
        seg_initial_delta_phi (ndarray): Array of initial_delta_phi values from seg_list
        seg_idpu_spinper (ndarray): Array of idpu_spinper values from seg_list
        seg_segflags (ndarray): Array of segflags values from seg_list
        seg_count (int): Count of segments

    """

    def print_segs(self):
        """ Print the list of SpinmodelSegment objects, useful for troubleshooting.

          """

        for seg in self.seg_list:
            seg.print()

    def adjust_delta_phi(self,
                         ecl_start: float,
                         ecl_end: float,
                         ecl_corr: float):
        """ Apply the level 2 (spin fit) corrections to the segments falling in an eclipse time interval.

        Args:
            ecl_start (float): Start time (seconds since epoch) of an eclipse to be processed
            ecl_end (float): End time (seconds since epoch) of an eclipse to be processed
            ecl_corr (float): Eclipse delta-phi offset (degrees) between waveform and spin fit data for this eclipse
          """

        # A segment needs to be updated if its midpoint lies in the time range ecl_start to ecl_end.
        # This is more robust against floating point roundoff errors than the previous method, which
        # compared the segment endpoints rather than the midpoints.

        seg_midpoints = (self.seg_times + self.seg_t2) / 2.0
        cond1 = ecl_start <= seg_midpoints
        cond2 = seg_midpoints <= ecl_end
        cond = cond1 & cond2
        idx = np.arange(len(self.seg_list))[cond]
        self.seg_initial_delta_phi[idx] += ecl_corr
        self.seg_segflags[idx] |= 4
        # print("Eclipse start: ", ecl_start, "Eclipse end: ", ecl_end)
        # print("First eclipse segment")
        # self.seg_list[idx[0]].print()
        # print("Last eclipse segment")
        # idx_len = len(idx)
        # self.seg_list[idx[idx_len - 1]].print()
        for si in idx:
            seg = self.seg_list[si]
            seg.initial_delta_phi += ecl_corr
            seg.segflags |= 4

    def findseg_t(self,
                  t: np.ndarray) -> np.ndarray:
        """ Return an ndarray of index values for the segments covering a set of input times

        This is a helper routine for the interp_t method.

        Args:
            t (ndarray(dtype=float): Input times

        Returns:
            ndarray(dtype=int) of indices into the seg_* attributes for each input time
         """
        n = len(self.seg_times)
        idx1 = np.searchsorted(self.seg_times, t, side="right") - 1
        idx1_clip = np.clip(idx1, 0, n-1)

        # idx1_clip is what will be returned.  We'll do a consistency check for any gaps or
        # overlaps between segment times, just in case any weird floating point stuff happens to the
        # seg_times or seg_t2 arrays, causing input times to fall in the cracks between segments, or match
        # multiple segments.

        idx2 = np.searchsorted(self.seg_t2, t, side="right") - 1
        idx2_clip = np.clip(idx2, -1, n-2)
        mismatch = ((idx2_clip + 1) != idx1_clip)
        bad_idx = np.flatnonzero(mismatch)
        c = bad_idx.size
        if c != 0:
            logging.warning('Indices don''t match up: time gaps or overlaps detected. Mismatched indices:')
            logging.warning(bad_idx)
            raise ValueError
        else:
            return idx1_clip

    def interp_t(self,
                 t: np.ndarray,
                 use_spinphase_correction: bool = True) -> SpinmodelInterpTResult:
        """ Interpolate the spin model to a set of input times, and return an object holding the results.

        This is the workhorse routine for accessing the spin model. Clients will specify a set of input times (e.g.
        from a tplot variable to be operated on), then obtain the spin status at those times, which can be used to
        transform back and forth between SSL coordinates (spinning with the spacecraft) and DSL (despun, referenced to
        sun direction and angular momentum vector/spin axis).

        Rather than inefficiently iterating over the spinmodel segments and input times, the code performs vectorized
        operations on Numpy arrays. This is done by calculating a set of segment indices corresponding to the input time
        stamps, then replicating the segment data to match the number of input times. (So, we are trading extra memory
        usage to allow vectorized bulk calculations)  To minimize branching, several boolean conditions are evaluated,
        yielding sets of array indices matching each set of conditions. The appropriate calculations are performed in
        bulk for each set of condition indices. Finally, the various outputs are collected into a SpinmodelInterpTResult
        object and returned.

        Args:
            t (ndarray(dtype=float): Input times
            use_spinphase_correction (Boolean): Flag (defaults to True) specifying whether V03 state CDF corrections
                should be applied to the interpolation output.

        Returns:
            SpinmodelInterpTResult object containing the interpolated outputs (spinphase, spincount, spin period,
            eclipse corrections, etc)
         """

        segs_idx = self.findseg_t(t)
        n = len(segs_idx)

        my_seg_times = self.seg_times[segs_idx]
        my_seg_t2 = self.seg_t2[segs_idx]
        my_seg_b = self.seg_b[segs_idx]
        my_seg_c = self.seg_c[segs_idx]
        my_seg_c1 = self.seg_c1[segs_idx]
        my_seg_c2 = self.seg_c2[segs_idx]
        my_seg_initial_delta_phi = self.seg_initial_delta_phi[segs_idx]
        my_seg_idpu_spinper = self.seg_idpu_spinper[segs_idx]
        my_seg_segflags = self.seg_segflags[segs_idx]

        # Below this point, all indexing assumes n = #times, not n = #model_segments.  No references to self.anything.
        all_idx = np.arange(n)
        # output variables
        spincount = np.zeros(n)
        t_last = np.zeros(n)
        spinphase = np.zeros(n)
        spinper = np.zeros(n)
        eclipse_delta_phi = np.zeros(n)  # It is important that this variable is initialized to 0
        mask = np.zeros(n) + 3

        # Internal variables.
        # If memory becomes an issue, we can avoid allocating by using some creative index operations
        dt = np.zeros(n)
        fracspins = np.zeros(n)
        intspins = np.zeros(n, int)
        bp = np.zeros(n)
        phi_lastpulse = np.zeros(n)
        tlast_dt = np.zeros(n)

        idx1_cond = t < my_seg_times
        idx1 = all_idx[idx1_cond]
        c1 = len(idx1)

        idx2_cond = t > my_seg_t2
        idx2 = all_idx[idx2_cond]
        c2 = len(idx2)

        # each of the conditions below splits based on the truth of this subclause
        branch1_cond = np.abs(my_seg_c) < 1.0e-12
        branch1_idx = all_idx[branch1_cond]
        branch1_cidx = all_idx[~branch1_cond]
        branch1_c = len(branch1_idx)
        branch1_nc = len(branch1_cidx)
        tmp1 = np.bitwise_and(my_seg_segflags, 3)
        tmp2 = np.equal(tmp1, mask)
        tmp3 = np.greater(my_seg_idpu_spinper, 1.0)
        branch2_cond = tmp2 & tmp3
        branch2_idx = all_idx[branch2_cond]

        idx3_cond = (t >= my_seg_times) & (t <= my_seg_t2)
        idx3 = all_idx[idx3_cond]
        c3 = len(idx3)

        if c1 > 0:
            dt[idx1] = my_seg_times[idx1] - t[idx1]
            spinper[idx1] = 360.0 / my_seg_b[idx1]
            fracspins[idx1] = dt[idx1] / spinper[idx1]
            intspins[idx1] = np.ceil(fracspins[idx1])
            spinphase[idx1] = (intspins[idx1] - fracspins[idx1]) * 360.0
            spincount[idx1] = my_seg_c1[idx1] - intspins[idx1]
            t_last[idx1] = my_seg_times[idx1] - intspins[idx1] * spinper[idx1]

        if c2 > 0:
            dt[idx2] = t[idx2] - my_seg_t2[idx2]
            bp[idx2] = my_seg_b[idx2] + 2.0 * my_seg_c[idx2] * (my_seg_t2[idx2] - my_seg_times[idx2])
            spinper[idx2] = 360.0 / bp[idx2]
            fracspins[idx2] = dt[idx2] / spinper[idx2]

            idx2_branch_cond = idx2_cond & branch2_cond
            idx2_branch = all_idx[idx2_branch_cond]
            if len(idx2_branch) > 0:
                model_phi = fracspins[idx2_branch] * 360.0
                idpu_bp = 360.0 / my_seg_idpu_spinper[idx2_branch]
                idpu_phi = dt[idx2_branch] * idpu_bp
                eclipse_delta_phi[idx2_branch] = my_seg_initial_delta_phi[idx2_branch] + (model_phi - idpu_phi)
            intspins[idx2] = np.floor(fracspins[idx2])
            spinphase[idx2] = (fracspins[idx2] - intspins[idx2]) * 360.0
            spincount[idx2] = my_seg_c2[idx2] + intspins[idx2]
            t_last[idx2] = my_seg_t2[idx2] + intspins[idx2] * spinper[idx2]

        if c3 > 0:
            dt[idx3] = t[idx3] - my_seg_times[idx3]
            phi = my_seg_b[idx3] * dt[idx3] + my_seg_c[idx3] * dt[idx3] * dt[idx3]
            bp[idx3] = my_seg_b[idx3] + 2.0 * my_seg_c[idx3] * dt[idx3]
            spinper[idx3] = 360.0 / bp[idx3]
            spinphase[idx3] = np.fmod(phi, 360.0)
            fracspins[idx3] = phi / 360.0
            spincount[idx3] = np.floor(fracspins[idx3])
            phi_lastpulse[idx3] = spincount[idx3] * 360.0

            if branch1_c > 0:
                tlast_dt[branch1_idx] = phi_lastpulse[branch1_idx] / my_seg_b[branch1_idx]

            if branch1_nc > 0:
                tlast_dt[branch1_cidx] = (-my_seg_b[branch1_cidx] +
                                          np.sqrt(my_seg_b[branch1_cidx] ** 2 - 4.0 * my_seg_c[branch1_cidx] * (
                                              -phi_lastpulse[branch1_cidx]))) / (2.0 * my_seg_c[branch1_cidx])

            idx3_branch_cond = idx3_cond & branch2_cond
            idx3_branch = all_idx[idx3_branch_cond]

            if len(idx3_branch) > 0:
                model_phi = fracspins[idx3_branch] * 360.0
                idpu_bp = 360.0 / my_seg_idpu_spinper[idx3_branch]
                idpu_phi = dt[idx3_branch] * idpu_bp
                eclipse_delta_phi[idx3_branch] = my_seg_initial_delta_phi[idx3_branch] + (model_phi - idpu_phi)
            t_last[idx3] = my_seg_times[idx3] + tlast_dt[idx3]
            spincount[idx3] = spincount[idx3] + my_seg_c1[idx3]

        if use_spinphase_correction:
            logging.info("applying spinphase correction")
            interp_correction = np.interp(t, self.spin_corr_times, self.spin_corr_vals)
            spinphase -= interp_correction
            cond = spinphase > 360.0
            spinphase[cond] -= 360.0
            cond = spinphase < 0.0
            spinphase[cond] += 360.0

        res = SpinmodelInterpTResult(spincount=spincount, spinphase=spinphase, t_last=t_last,
                                     eclipse_delta_phi=eclipse_delta_phi, spinper=spinper, segflags=my_seg_segflags,
                                     idx=segs_idx, dt=dt)
        return res

    def make_arrays(self):
        """ Populate the seg_* attributes using data from the segment list

         For compatibility with the IDL implementation, we use the same algorithm in Python to build up a segment
         list from the STATE CDF variables.   In IDL, we can do vectorized access to class members through an array
         of indices, but this doesn't work the same way in Python.  Here, it is most convenient to convert the segment
         list to a set of Numpy arrays, which lets us do vectorized calculations on them.

         Args:

         Returns:
           """

        self.seg_count = len(self.seg_list)
        self.seg_times = np.array([o.t1 for o in self.seg_list])
        self.seg_t2 = np.array([o.t2 for o in self.seg_list])
        self.seg_c1 = np.array([o.c1 for o in self.seg_list])
        self.seg_c2 = np.array([o.c2 for o in self.seg_list])
        self.seg_b = np.array([o.b for o in self.seg_list])
        self.seg_c = np.array([o.c for o in self.seg_list])
        self.seg_npts = np.array([o.npts for o in self.seg_list])
        self.seg_maxgap = np.array([o.maxgap for o in self.seg_list])
        self.seg_phaserr = np.array([o.phaserr for o in self.seg_list])
        self.seg_idpu_spinper = np.array([o.idpu_spinper for o in self.seg_list])
        self.seg_initial_delta_phi = np.array([o.initial_delta_phi for o in self.seg_list])
        self.seg_segflags = np.array([o.segflags for o in self.seg_list])

    def make_tplot_vars(self,
                        prefix: str):
        """ Create a set of tplot variables from the spinmodel segment attributes.

        This is useful for regression testing or cross-platform validation of the spin model creation process.

        :param prefix : A string to prepend to each tplot variable name to ensure uniqueness
        :return:
        """

        store_data(prefix + 't1', data={'x': self.seg_times, 'y': self.seg_times})
        store_data(prefix + 't2', data={'x': self.seg_times, 'y': self.seg_t2})
        store_data(prefix + 'c1', data={'x': self.seg_times, 'y': self.seg_c1})
        store_data(prefix + 'c2', data={'x': self.seg_times, 'y': self.seg_c2})
        store_data(prefix + 'b', data={'x': self.seg_times, 'y': self.seg_b})
        store_data(prefix + 'c', data={'x': self.seg_times, 'y': self.seg_c})
        store_data(prefix + 'npts', data={'x': self.seg_times, 'y': self.seg_npts})
        store_data(prefix + 'maxgap', data={'x': self.seg_times, 'y': self.seg_maxgap})
        store_data(prefix + 'phaserr', data={'x': self.seg_times, 'y': self.seg_phaserr})
        store_data(prefix + 'idpu_spinper', data={'x': self.seg_times, 'y': self.seg_idpu_spinper})
        store_data(prefix + 'initial_delta_phi', data={'x': self.seg_times, 'y': self.seg_initial_delta_phi})
        store_data(prefix + 'segflags', data={'x': self.seg_times, 'y': self.seg_segflags})

    def addseg(self,
               newseg: SpinmodelSegment):
        """ Add a segment to the spin model object being constructed.

        A spin model is assumed to satisfy a condition that the end time of one segment exactly matches the start time
        of the next segment, with no gaps or overlaps.

        When loading STATE data for time intervals spanning multiple UTC days, there will be small time gaps
        between the spin model segments at either side of UTC date boundaries.  This routine adjusts the segment
        list by inserting new 'bridge' segments, as needed, to bridge gaps between the segments read from the CDF data,
        also making any necessary adjustments to the preceding/following segments.

        :param newseg: A SpinmodelSegment object to be added to the spin model being constructed
        :return:
        """

        if self.seg_count == 0:
            self.seg_list.append(newseg)
            self.lastseg = 0
            self.seg_count = 1
        else:
            lseg = self.seg_list[self.lastseg]
            # print('Adding segment')
            # print('Last:')
            # lseg.print()
            # print('Current')
            # newseg.print()
            # Previously, this was an exact equality test.  Now, the DEPEND_TIME variables are handled slightly
            # differently than plain old double-precision variables, so the possibility of small floating point
            # differences needs to be accounted for. For the purposes of spin model segments, if the times are within
            # a microsecond, they might as well be equal.
            tdiff = abs(newseg.t1 - lseg.t2)
            tolerance = 1.0e-06
            if tdiff < tolerance:

                #
                # Normal case: segments are contiguous
                #
                newseg.c1 = lseg.c2
                newseg.c2 = lseg.c2 + newseg.c2
                self.seg_list.append(newseg)
                self.lastseg += 1
                self.seg_count += 1
            else:
                # Segments are not contiguous -- this should indicate
                # a UTC date boundary, and the spin models on either side will
                # need to be merged.
                #
                # There are several cases, depending on the delta-t between the
                # end of the previous segment, and the start of the current segment:
                #
                # 1) Large gap, greater than 1/2 spin : create a new segment to
                #    bridge the gap.
                # 2) Small gap, <= 1/2 spin, previous segment covers 2 or more spins:
                #    remove last spin from previous segment, converting the situation
                #    to the "large gap" case, then create a new segment to bridge
                #    the gap.
                # 3) Small gap, previous segment only contains 1 spin : if current
                #    segment contains 2 or more spins, remove first spin from
                #    current segment, converting the situation to the "large gap"
                #    case, then create a new segment to bridge the gap.
                # 4) Small gap, previous and current segments each contain only
                #    a single spin.  This should never happen -- if no averaging
                #    was applied, the segments should be exactly contiguous.
                # 5) Negative gap -- current segment starts more than 1/2 spin
                #    before end of previous segment.  This should never happen,
                #    since it would imply that the apid 305 packets are incorrectly
                #    time ordered.

                spinper = 360.0 / lseg.b
                gap_spin_count = (newseg.t1 - lseg.t2) / spinper
                gap_time = newseg.t1 - lseg.t2
                if gap_spin_count > 0.5:
                    # Case 1: Gap of 1 or more spins between segments, add fill
                    gap_nspins = math.floor(gap_spin_count + 0.5)
                    gap_spinper = (newseg.t1 - lseg.t2) / (1.0 * gap_nspins)
                    # Fill in eclipse delta_phi parameters
                    gap_idpu_spinper = (newseg.idpu_spinper + lseg.idpu_spinper) / 2.0
                    gap_segflags = newseg.segflags & lseg.segflags
                    # We need to calculate gap_initial_delta_phi by extrapolating
                    # from lseg to lseg.t2 = gapseg.t1
                    spinper, spinphase, dummy_spincount, gap_eclipse_delta_phi, dummy_t_last = \
                        lseg.interp_t(lseg.t2)
                    fillseg = SpinmodelSegment(t1=lseg.t2, t2=newseg.t1, c1=lseg.c2, c2=lseg.c2 + gap_nspins,
                                               b=360.0 / gap_spinper, c=0.0, npts=0, maxgap=gap_time, phaserr=0.0,
                                               initial_delta_phi=gap_eclipse_delta_phi,
                                               idpu_spinper=gap_idpu_spinper, segflags=gap_segflags)
                    self.seg_list.append(fillseg)
                    self.lastseg = self.lastseg + 1
                    self.seg_count += 1
                    newseg.c1 = fillseg.c2
                    newseg.c2 = newseg.c1 + newseg.c2
                    self.seg_list.append(newseg)
                    self.lastseg = self.lastseg + 1
                    self.seg_count += 1
                elif gap_spin_count > -0.5:
                    # Case 2, 3, 4, or 5
                    # Now that we're using floating point time comparisons with tolerance, rather than strict
                    # equality, none of these cases seem to occur anymore. Cases 2 and (possibly) 3 have apparently
                    # been absorbed into the "segments are contiguous (within tolerance)" case.  Cases 4 and 5 were
                    # "this should never happen" scenarios.  It would be somewhat difficult to contrive plausible test
                    # inputs that would trigger cases 2 or 3 in this iteration of the code, so (for the sake of our test
                    # coverage metrics), I will replace those code blocks with exceptions. If they ever do turn up
                    # in practice, we can revive them from the version history or from the IDL code.  -- JWL 2023/03/13

                    if (lseg.c2 - lseg.c1) >= 2:
                        # Case 2: small gap, previous segment has at least 2 spins
                        # dprint,'<1 spin gap, stealing spin from last segment'
                        logging.error('Unexpected case 2 (small gap, previous segment with at least 2 spins)')
                        logging.error('Segment time: ' + time_string(lseg.t2))
                        logging.error('Please contact pyspedas or themis support and include the above information.')
                        raise RuntimeError
                    elif newseg.c2 >= 2:
                        # Case 3: small gap, previous segment has only 1 spin, current segment has at least 2 spins
                        # It is assumed that newseg is the first segment of a new UTC day, therefore the spin numbers
                        # start over at 0.  So we want to change newseg to start at spin 1 instead of spin 0.
                        logging.error('Unexpected case 3 (small gap, previous segment with only 1 spin, current segment with 2+ spins.)')
                        logging.error('Segment time: ' + time_string(lseg.t2))
                        logging.error('Please contact pyspedas or themis support and include the above information.')
                        raise RuntimeError
                    else:
                        # Case 4: small gap, but segments on either side only contain
                        # one spin each.  This should never happen.
                        logging.error('Unexpected case 4 (<1 spin gap, but neither segment has enough spins to steal.)')
                        logging.error('Segment time: ' + time_string(lseg.t2))
                        logging.error('Please contact pyspedas or themis support and include the above information.')
                        raise RuntimeError
                else:
                    # Case 5: out of order sun pulse times.  This should never happen.
                    logging.error('Unexpected case 5 (Sun pulse times out of order)')
                    logging.error("Last segment end time" + time_string(lseg.t2) + " New segment start time " + time_string(newseg.t1))
                    raise RuntimeError

    def get_timerange(self):
        """ Returns the time span covered by the model.
        The IDL version also returns information about any eclipse time periods, will add later
        if needed.

        Args: None

        Returns: tuple with start time and end time
        """
        start_time = self.seg_times[0]
        end_time = self.seg_t2[-1]
        return start_time, end_time

    def get_eclipse_times(self):
        """ Returns lists of start time and end times for eclipses found in the spin model

        Returns: A tuple containing two lists, one for start times and one for end times.  Returns empty
            lists if no eclipses found.
        """
        start_times=[]
        end_times=[]
        processing_shadow=False

        for i in range(self.seg_count):
            this_eclipse_flag=self.seg_segflags[i] & 1
            if not this_eclipse_flag and not processing_shadow:
                # Previous and current segments are not eclipses, do nothing
                pass
            elif not this_eclipse_flag and processing_shadow:
                # Transition out of shadow, reset status
                processing_shadow=False
            elif this_eclipse_flag and not processing_shadow:
                # Transition into shadow, add entries to start and end time lists, set status
                start_times.append(self.seg_times[i])
                end_times.append(self.seg_t2[i])
                processing_shadow=True
            else:
                # Previous and current segments in shadow, update last end time
                end_times[-1]=self.seg_t2[i]

        return start_times, end_times


    def __init__(self,
                 probe,
                 correction_level,
                 suffix=''):
        self.lastseg = -1
        self.seg_times = np.zeros(1, float)
        self.seg_t2 = np.zeros(1, float)
        self.seg_c1 = np.zeros(1, int)
        self.seg_c2 = np.zeros(1, int)
        self.seg_b = np.zeros(1, float)
        self.seg_c = np.zeros(1, float)
        self.seg_npts = np.zeros(1, int)
        self.seg_maxgap = np.zeros(1, float)
        self.seg_phaserr = np.zeros(1, float)
        self.seg_initial_delta_phi = np.zeros(1, float)
        self.seg_idpu_spinper = np.zeros(1, float)
        self.seg_segflags = np.zeros(1, int)
        self.seg_count = 0
        self.seg_list = []
        seg_times, tend_data = get_sm_data(probe, 'tend', correction_level, suffix)
        _, spinper_data = get_sm_data(probe, 'spinper', correction_level, suffix)
        _, c_data = get_sm_data(probe, 'c', correction_level, suffix)
        _, phaserr_data = get_sm_data(probe, 'phaserr', correction_level, suffix)
        _, nspins_data = get_sm_data(probe, 'nspins', correction_level, suffix)
        _, npts_data = get_sm_data(probe, 'npts', correction_level, suffix)
        _, maxgap_data = get_sm_data(probe, 'maxgap', correction_level, suffix)
        _, initial_delta_phi_data = get_sm_data(probe, 'initial_delta_phi', correction_level, suffix)
        _, idpu_spinper_data = get_sm_data(probe, 'idpu_spinper', correction_level, suffix)
        _, segflags_data = get_sm_data(probe, 'segflags', correction_level, suffix)
        # The spin_correction variable only exists in V03 state CDFs, and has its own time variable
        tmp = get_sm_data(probe, 'correction', 0, suffix)
        if tmp is None:
            logging.info('spin_correction variable not available, defaulting to 0.0')
            self.spin_corr_times = [0.0, 1.0]
            self.spin_corr_vals = [0.0, 0.0]
        else:
            self.spin_corr_times, self.spin_corr_vals = tmp

        # The fgm_corr_offset and fgm_corr_tend variables may not exist, and have their own time variable
        tmp = get_sm_data(probe, 'fgm_corr_offset', correction_level, suffix)
        if tmp is None:
            do_fgm_corr = False
            logging.info('FGM correction variables not available')
        else:
            do_fgm_corr = True
            fgm_corr_time, fgm_corr_offset = tmp
            _, fgm_corr_tend = get_sm_data(probe, 'fgm_corr_tend', correction_level, suffix)

        seg_count = len(seg_times)
        # tlast = seg_times[0]
        for i in range(seg_count):
            newseg = SpinmodelSegment(t1=seg_times[i], t2=tend_data[i], c1=0, c2=nspins_data[i],
                                      b=360.0 / spinper_data[i],
                                      c=c_data[i], npts=npts_data[i], maxgap=maxgap_data[i], phaserr=phaserr_data[i],
                                      initial_delta_phi=initial_delta_phi_data[i], idpu_spinper=idpu_spinper_data[i],
                                      segflags=segflags_data[i])
            # tlast=tend_data[i]
            # origseg=copy.copy(newseg)
            self.addseg(newseg)
            # print(i)
            # newseg.print()
            # lastseg=copy.copy(origseg)
        self.make_arrays()
        if do_fgm_corr and (correction_level == 2):
            logging.info(f"applying FGM corrections, do_fgm_corr = {do_fgm_corr}, correction_level = {correction_level}")
            for i in np.arange(len(fgm_corr_offset)):
                self.adjust_delta_phi(fgm_corr_time[i], fgm_corr_tend[i], fgm_corr_offset[i])
        else:
            logging.info(f"Skipping FGM corrections, do_fgm_corr {do_fgm_corr}, correction_level = {correction_level}")


# This dictionary is where the spinmodel objects are stored.   The keys are tuples of (probe, correction_level)
# and the values are Spinmodel objects.  Spinmodel objects are added to the dictionary by the
# spinmodel_postprocess routine.

spinmodel_dict: Dict[(str, int)] = {}


def get_spinmodel(probe: str,
                  correction_level: int,
                  quiet: bool = False) -> Spinmodel:
    """ Get a reference to a Spinmodel object stored in the dictionary.

    Args:
        probe: Probe name, one of 'a','b','c','d','e'
        correction_level: 0 for no corrections, 1 for waveform corrections, 2 for spin fit corrections
        quiet:  If True, do not log anything if the model is uninitialized

    Returns:
        A reference to a Spinmodel object stored in the dictionary.
    """
    try:
        model = spinmodel_dict[(probe, correction_level)]
    except KeyError:
        if not quiet:
            logging.warning("No spinmodel loaded for probe " + probe + " correction level: " + str(correction_level))
            logging.warning("It is necessary to load THEMIS state data, with get_support_data=True, to initialize the spin model.")
        model = None
    return model


def save_spinmodel(probe: str,
                   correction_level: int,
                   model: Spinmodel):
    """ Store a reference to a Spinmodel object in the dictionary, using the probe and correction level as the key

    Args:
        probe: Probe name, one of 'a','b','c','d','e'
        correction_level: 0 for no corrections, 1 for waveform corrections, 2 for spin fit corrections
        model: A Spinmodel object to store

    """
    spinmodel_dict[(probe, correction_level)] = model



import logging
import numpy as np
import math
from geopack.geopack import recalc as geopack_recalc
from pyspedas import cotrans, time_double


def sm_ns_model(time, gsm_pos, sc2NS=False):
    """
    This routine calculates the NS position along the zaxis at a specific x and y location.

    Parameters
    ----------
    time :  array_like of float or string
        Time array.
    gsm_pos :  array_like of float
        Data array in GSM coordinates.
    sc2NS : bool, optional
        If True, return SM coordinates along z-axis.
        If False, return the difference of gms to sm along the z-axis.
        Default is False.

    Returns
    -------
    type
        Description of returned object.

    """
    # convert gsm to sm coordinates
    sm_pos = cotrans(time_in=time, data_in=gsm_pos, coord_in="gsm", coord_out="sm")

    if sc2NS:
        return gsm_pos[:, 2]
    return gsm_pos[:, 2] - sm_pos[:, 2]


def themis_ns_model(time, gsm_pos, sc2NS=False):
    """
    This routine calculates the position along the zaxis at a specific x and y location. The themis model is used for this calculation.
    The themis model uses z-sm (converted from z-gsm) for the inner probes and the Hammond model for the outer probes.

    Parameters
    ----------
    time : array_like of str or float
        Time in either string or double format.
        If float, represents seconds since 1970.
        If string, should be in the format: "YYYY-MM-DD/hh:mm:ss".
    gsm_pos : array_like of float
        Position vector in GSM coordinates in re (pos[*,3]).
    sc2NS : bool, optional
        If set, returns Z displacement from the spacecraft to the neutral sheet.
        Value is positive if the NS is northward of the SC location, and negative if below. Default is False.

    Returns
    -------
    list of float
        Returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS).
        Value is positive if NS is above z=0 gsm plane, negative if below.

    Notes
    -----
    The themis model uses z-sm (converted from z-gsm) for the inner probes and the Hammond model (default) for the outer probes.
    The algorithm can be found in ssllib neutralsheet.pro.
    """
    # initialize constants and variables
    re = 6378.0
    h0 = 8.6  # 10.5 # hinge point of the neutral sheet
    rad = np.pi / 180.0
    dz2NS = np.zeros(len(time))

    # constants used in hammond model
    H1 = 8.6
    Y0 = 20.2
    D = 12.2

    # calculate the radial distance
    rdist = np.sqrt(gsm_pos[:, 0] ** 2 + gsm_pos[:, 1] ** 2 + gsm_pos[:, 2] ** 2)

    # Use the sm coordinates for radial distances <= h0  (8.6)
    sm_ind = np.argwhere(rdist <= h0).flatten()
    if len(sm_ind) > 0:
        sm_pos = cotrans(
            time_in=time[sm_ind],
            data_in=gsm_pos[sm_ind, :],
            coord_in="gsm",
            coord_out="sm",
        )
        dz2NS[sm_ind] = -sm_pos[:, 2]

    # Use the Hammond model for radial distances > h0  (8.6)
    lr_ind = np.argwhere(rdist > h0).flatten()
    if len(lr_ind) > 0:
        # initialize variables
        x = gsm_pos[lr_ind, 0]
        y = gsm_pos[lr_ind, 1]
        z = gsm_pos[lr_ind, 2]
        tilt = np.zeros(len(x))
        # check input time format and convert to doy, hr, min
        for i in range(len(x)):
            # calculate the tilt in degrees
            tilt[i] = geopack_recalc(time_double(time[lr_ind[i]]))

        # hammond model
        iless = np.argwhere(np.abs(y) < Y0).flatten()
        if len(iless) > 0:
            dz2NS[lr_ind[iless]] = (
                (H1 + D) * np.sqrt(1 - y[iless] ** 2 / Y0**2) - D
            ) * np.sin(tilt[iless])

        imore = np.argwhere(np.abs(y) >= Y0).flatten()
        if len(imore) > 0:
            dz2NS[lr_ind[imore]] = -D * np.sin(tilt[imore])

    if not sc2NS:
        return gsm_pos[:, 2] - (-dz2NS)
    else:
        return -dz2NS


def aen_ns_model(time, gsm_pos, sc2NS=False):
    """
    Find the AEN (Analytical Equatorial Neutral) sheet in the magnetopause in different time and position.

    Parameters
    ----------
    time : array_like of str or float
        Time in either string or double format.
        If float, represents seconds since 1970.
        If string, should be in the format: "YYYY-MM-DD/hh:mm:ss".
    gsm_pos : array_like
        Position vector in GSM coordinates in re (pos[*,3]).
    sc2NS : bool, optional
        If set, returns Z displacement from the spacecraft to the neutral sheet.
        Value is positive if the NS is northward of the SC location, and negative if below. Default is False.

    Returns
    -------
    list of float
        Returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS).
        Value is positive if NS is above z=0 gsm plane, negative if below.

    References
    ----------
    (1) AEN(Analytical Equatorial Neutral):
        Zhu, M. and R.-L. Xu, 1994, A continuous neutral sheet model and a normal
        curved coordinate system in the magnetotail,  Chinese J. Space Science,  14,
        (4)269, (in Chinese).
        Wang, Z.-D. and R.-L. Xu, Neutral Sheet Observed on ISEE Satellite,
        Geophysical Research Letter, 21, (19)2087, 1994.
    (2) Magnetopause model:
        Sibeck, D. G., R. E. Lopez, and E. C. Roelof, Solar wind control of the
        magnetopause shape, location, and motion, J. Grophys. Res., 96, 5489, 1991
    """
    # initialize constants
    h0 = 12.6 / np.pi

    dz2ns = np.zeros(len(time))

    for i in range(len(time)):
        # calculate the tilt angle
        tt = geopack_recalc(time_double(time[i]))

        # calculate the position of the neutral sheet
        dz2ns[i] = (
            -h0
            * np.sin(tt)
            * np.arctan(gsm_pos[i, 0] / 5)
            * (2 * np.cos(gsm_pos[i, 1] / 6))
        )

    if not sc2NS:
        return dz2ns
    else:
        return gsm_pos[:, 2] - dz2ns


def den_ns_model(time, gsm_pos, sc2NS=False):
    """
    Find the DEN (Displaced Equatorial Neutral) sheet inside the magnetopause in different time and positions.
    The routine calculates the position along the zaxis at a specific location.

    Parameters
    ----------
    time : array_like of str or float
        Time in either string or double format.
        If float, represents seconds since 1970.
        If string, should be in the format: "YYYY-MM-DD/hh:mm:ss".
    gsm_pos : array_like
        Position vector in GSM coordinates in re (pos[*,3]).
    sc2NS : bool, optional
        If set, returns Z displacement from the spacecraft to the neutral sheet.
        Value is positive if the NS is northward of the SC location, and negative if below. Default is False.

    Returns
    -------
    float
        Returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS).
        Value is positive if NS is above z=0 gsm plane, negative if below.

    References
    ----------
    (1) DEN(Displaced Equatorial Neutral):
        Xu, R.-L., A Displaced Equatorial Neutral Sheet Surface Observed on ISEE-2
        Satellite,  J. Atmospheric and Terrestrial Phys., 58, 1085, 1991
    (2) Magnetopause model:
        Sibeck, D. G., R. E. Lopez, and R. C. Roelof, Solar wind control of the
        magnetopause shape, location, and motion, J. Grophys. Res., 96, 5489, 1991

    Notes
    -----
    Original Authors of the FORTRAN source code:
    Ronglan XU and Lei LI, Center for Space Sci. and Applied Res.,
    Chinese Academy of Sciences, PO Box 8701, Beijing 100080, China
    E-mail: XURL@SUN.IHEP.AC.CN, XURL@SUN20.CSSAR.AC.CN
    This source code was ported from the original FORTRAN source code into IDL
    The original source code only calculated to 10.05 RE. In this IDL version
    that restriction was increased to 25.
    """

    # calculate the position of the neutral sheet along z axis
    H = 25.5
    H1 = 25.05

    dz2ns = np.zeros(len(time))

    for i in range(len(time)):
        done = 0
        xgsm = gsm_pos[i, 0]
        ygsm = gsm_pos[i, 1]

        # get tilt angle of magnetic pole
        tilt = geopack_recalc(time_double(time[i]))

        if xgsm > -100.0:
            d = sd1(tilt, H, H1, xgsm)
            ym21 = ((H1 * (H + d)) ** 2) * (1 - (xgsm / (H * np.cos(tilt))) ** 2)
            ym22 = (H + d) ** 2 - (d - xgsm / np.cos(tilt)) ** 2
            ym2 = ym21 / ym22
            if ym2 < 0:
                # ie[i] = 2
                continue
            ym = np.sqrt(ym2)
            xd2 = ((H * np.cos(tilt)) ** 2) * (1 - (ygsm / H1) ** 2)
            if np.abs(ygsm) > H1:
                xd2 = 0
            # find the equatorial region
            xd = np.sqrt(xd2)
            rd = np.sqrt(xd**2 + ygsm**2)
            rsm = np.sqrt(xgsm**2 + ygsm**2)
            if xgsm > 0 or rsm <= rd:
                dz2ns[i] = -xgsm * np.sin(tilt) / np.cos(tilt)
                done = 1
            if np.abs(ygsm) > ym and done != 1:
                dz2ns[i] = -d * np.sin(tilt)
                done = 1
            if done != 1:
                dz2ns[i] = ((H + d) * np.sqrt(1 - (ygsm**2) / ym2) - d) * np.sin(tilt)

    if not sc2NS:
        return dz2ns
    else:
        sc2NS = gsm_pos[:, 2] - dz2ns
        return sc2NS


def sfa4(aa, bb, cc, dd):
    ndx = 0
    xmin = 0
    xmax = 50
    ndxmax = 3
    dx = 1
    x = xmin
    yy = x**4 + aa * x**3 + bb * x**2 + cc * x + dd

    while ndx <= ndxmax:
        x = x + dx
        if x >= xmax:
            ndx = 0
            return x
        y = x**4 + aa * x**3 + bb * x**2 + cc * x + dd
        ry = y / yy
        if ry < 0:
            x = x - dx
            dx = dx / 10.0
            ndx = ndx + 1
        else:
            yy = y
    return x


def sd1(til, H, H1, xgsm):
    ct = np.cos(til)
    xx = xgsm
    xh = -H * ct
    if xgsm >= xh:
        xx = xh

    # calculate the radius of the cross section
    if xx <= -5.0:
        rm = 9 * (10 - 3 * xx) / (10 - xx) + 3
    if xx > -5.0:
        rm = np.sqrt(18**2 - (xx + 5) ** 2)
    rm2 = rm**2

    # in cross_section areas above and below the neutral
    # sheet
    aa = 4 * H - (32 * rm2 * H**2) / (np.pi**2 * H1**2 * (H - xx / ct))
    bb = 2 * H**2 * (3.0 - 8.0 * rm2 / (np.pi**2 * H1**2))
    cc = 4 * (H**3)
    dd = H**4
    x = sfa4(aa, bb, cc, dd)

    d = x
    if xgsm >= xh:
        fk = -x / np.sqrt(-xh)
        d = -fk * np.sqrt(-xgsm)
    return d


def fairfield_ns_model(time, gsm_pos, sc2NS=False):
    """
    Calculate the position along the zaxis at a specific x and y location.
    The Fairfield model is used for this calculation.

    Parameters
    ----------
    time : array_like of str or float
        Time in either string or double format.
        If float, represents seconds since 1970.
        If string, should be in the format: "YYYY-MM-DD/hh:mm:ss".
    gsm_pos : array_like
        Position vector in GSM coordinates in re (pos[*,3]).
    sc2NS : bool, optional
        If set, returns Z displacement from the spacecraft to the neutral sheet.
        Value is positive if the NS is northward of the SC location, and negative if below. Default is False.

    Returns
    -------
    list of float
        Returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS).
        Value is positive if NS is above z=0 gsm plane, negative if below.

    References
    ----------
    A statistical determination of the shape and position of the
    geomagnetic neutral sheet,  Journal of Geophysical Research,
    Vol. 85, No A2, pages 775-780, February 1, 1980
    Author - D. Fairfield
    """

    # constants (in re)
    h0 = 10.5
    y0 = 22.5
    d = 14.0

    dz2NS = np.zeros(len(time))
    tilt = np.zeros(len(time))

    for i in range(len(time)):
        # calculate tilt angle of geomagnetic axis
        tilt[i] = geopack_recalc(time_double(time[i]))

    # calculate the position of the neutral sheet along z axis
    y_ge_y0 = np.argwhere(np.abs(gsm_pos[:, 0]) >= y0).flatten()
    y_lt_y0 = np.argwhere(np.abs(gsm_pos[:, 0]) < y0).flatten()
    if len(y_ge_y0) > 0:
        dz2NS[y_ge_y0] = -d * np.sin(tilt[y_ge_y0])
    if len(y_lt_y0) > 0:
        dz2NS[y_lt_y0] = (
            (h0 + d) * np.sqrt(1 - gsm_pos[y_lt_y0, 0] ** 2 / y0**2) - d
        ) * np.sin(tilt[y_lt_y0])

    if not sc2NS:
        return dz2NS
    else:
        sc2NS = gsm_pos[:, 2] - dz2NS
        return sc2NS


def den_fairfield_ns_model(time, gsm_pos, sc2NS=False):
    """
    Calculate the position along the zaxis at a specific x and y location.

    Parameters
    ----------
    time : array_like of str or float
        Time in either string or double format.
        If float, represents seconds since 1970.
        If string, should be in the format: "YYYY-MM-DD/hh:mm:ss".
    gsm_pos : array_like
        Position vector in GSM coordinates in re (pos[*,3]).
    sc2NS : bool, optional
        If set, returns Z displacement from the spacecraft to the neutral sheet.
        Value is positive if the NS is northward of the SC location, and negative if below. Default is False.

    Returns
    -------
    list of float
        Returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS).
        Value is positive if NS is above z=0 gsm plane, negative if below.
    """
    # initialize constants
    dz2ns = np.zeros(len(time))

    # Use the den model for radial distances <12.re
    rdist = np.sqrt(gsm_pos[:, 0] ** 2 + gsm_pos[:, 1] ** 2 + gsm_pos[:, 2] ** 2)
    sm_ind = np.argwhere(rdist <= 10.0).flatten()
    if len(sm_ind) > 0:
        dz2ns[sm_ind] = den_ns_model(time[sm_ind], gsm_pos[sm_ind, :])

    # use the fairfield model for radial distances >12.re
    lr_ind = np.argwhere(rdist > 10.0).flatten()
    if len(lr_ind) > 0:
        dz2ns[lr_ind] = fairfield_ns_model(time[lr_ind], gsm_pos[lr_ind, :])

    if not sc2NS:
        return dz2ns
    else:
        sc2NS = gsm_pos[:, 2] - dz2ns
        return sc2NS


def rthph2xyz(r, th, ph):
    """
    Helper function for the Lopez model.
    Converts spherical to cartesian coordinates.

    Parameters
    ----------
    r : float
        Radial distance.
    th : float
        Latitude in degrees (range: -90 to 90).
    ph : float
        Longitude in degrees.

    Returns
    -------
    tuple
        Cartesian coordinates (x, y, z).

    Notes
    -----
    th, ph are in degrees, and th is latitude (not colatitude) (i.e. [-90->90]).
    """
    FLAG = 6.8792e28
    FLAG98 = 0.98 * FLAG
    PI = 3.1415926535898

    thrad = th * PI / 180.0
    phrad = ph * PI / 180.0
    sth = np.sin(thrad)
    cth = np.cos(thrad)
    sph = np.sin(phrad)
    cph = np.cos(phrad)
    x = r * cth * cph
    y = r * cth * sph
    z = r * sth

    iflags = np.argwhere((r > FLAG98) | (th > FLAG98) | (ph > FLAG98)).flatten()
    if len(iflags) > 0:
        x[iflags] = FLAG
        y[iflags] = FLAG
        z[iflags] = FLAG

    return x, y, z


def lopez_ns_model(time, gsm_pos, kp=None, mlt=None, sc2NS=False):
    """
    Calculate the position along the zaxis at a specific x and y location.
    The Lopez model is used for this calculation.

    Parameters
    ----------
    time : array_like of str or float
        Time in either string or double format.
        If float, represents seconds since 1970.
        If string, should be in the format: "YYYY-MM-DD/hh:mm:ss".
    gsm_pos : array_like
        Position vector in GSM coordinates in re (pos[*,3]).
    kp : float, optional
        Kp index value.
    mlt : float, optional
        Magnetic local time in degrees (0=midnight).
    sc2NS : bool, optional
        If set, returns Z displacement from the spacecraft to the neutral sheet.
        Value is positive if the NS is northward of the SC location, and negative if below. Default is False.

    Returns
    -------
    array_like of float
        Returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS).
        Value is positive if NS is above z=0 gsm plane, negative if below.

    References
    ----------
    The position of the magnetotail neutral sheet in the near-Earth Region,
    Geophysical Research Letters, Vol. 17, No 10, pages 1617-1620, 1990
    Author - Ramon E. Lopez
    The lopez model is best used for distances <8.8 RE
    """
    # constants
    rad = np.pi / 180.0
    if kp is None:
        kp = 0
    if mlt is None:
        mlt = 0.0
    tilt = np.zeros(len(time))

    for i in range(len(time)):
        # calculate tilt angle of geomagnetic axis
        tilt[i] = geopack_recalc(time[i])

    # calculate the position of the neutral sheet along z axis
    rdist = np.sqrt(gsm_pos[:, 0] ** 2 + gsm_pos[:, 1] ** 2 + gsm_pos[:, 2] ** 2)
    mlat = (
        -(0.14 * kp + 0.69)
        * ((np.cos(rad * mlt)) ** 0.3333333)
        * (0.065 * (rdist**0.8) - 0.16)
        * tilt
        * 180.0
        / np.pi
    )
    mlat = mlat + tilt * 180.0 / np.pi

    # convert magnetic latitude to position
    x, y, z = rthph2xyz(rdist, mlat, mlt)

    if not sc2NS:
        return z
    else:
        sc2NS = gsm_pos[:, 2] - z
        return sc2NS


def tag14_equat_sheet(xgsm, ygsm, zgsm, psi, pdyn, byimf, bzimf):
    """
    Calculates the GSM coordinates of a point of the TAG14 equatorial sheet.

    Parameters
    ----------
    xgsm : float
        X coordinate of a point inside the magnetosphere.
    ygsm : float
        Y coordinate of a point inside the magnetosphere.
    zgsm : float
        Z coordinate of a point inside the magnetosphere.
    psi : float
        Geodipole tilt angle in radians.
    pdyn : float
        Solar wind ram pressure in nanoPascals.
    byimf : float
        Y-GSM component of the IMF in nT (averaged over previous 30 minutes).
    bzimf : float
        Z-GSM component of the IMF in nT (averaged over previous 30 minutes).

    Returns
    -------
    tuple
        GSM coordinates (xgsm_s, ygsm_s, zgsm_s) of a point of the TAG14 equatorial sheet,
        located at the same geocentric distance R=sqrt(XGSM^2+YGSM^2+ZGSM^2)
        and lying in the same GSM meridian plane as the original point {XGSM,YGSM,ZGSM}.

    References
    ----------
    Tsyganenko, N. A., V. A. Andreeva, and E. I. Gordeev, Internally and externally induced deformations
    of the magnetospheric equatorial current as inferred from spacecraft data,
    Ann.Geophys., v.33, pp.1-11, doi:10.5194/angeo-33-1-2015, 2015.

    Notes
    -----
    Translated from FORTRAN via Chat-GPT4.
    Original FORTRAN source: https://geo.phys.spbu.ru/~tsyganenko/models/cs/TAG14.for
    """
    PI = 3.1415926539
    ERR = 1e-8

    # Constants
    (
        RH0,
        RH1,
        RH2,
        RH3,
        RH4,
        RH5,
        T0,
        T1,
        A00,
        A01,
        A02,
        A10,
        A11,
        A12,
        ALPHA0,
        DALPHA1,
        DALPHA2,
        DALPHA3,
        XAPPA,
        BETA0,
        BETA1,
    ) = (
        11.01595255,
        6.054799445,
        0.8376136340,
        -2.283371357,
        -0.2505011095,
        -0.9648465280,
        0.2865647076,
        0.1847283161,
        2.909423481,
        -0.1585372900,
        0.5648121327,
        1.889466989,
        0.06086006001,
        0.4853873874,
        7.128509839,
        4.867577218,
        -0.2248219212,
        -0.1441988549,
        -0.2868279445,
        2.179888855,
        0.4035940369,
    )

    xgsm1 = xgsm
    ygsm1 = ygsm
    zgsm1 = zgsm
    r = math.sqrt(xgsm1**2 + ygsm1**2 + zgsm1**2)
    theta = math.acos(zgsm1 / r)

    if xgsm1 == 0 and ygsm1 == 0:
        phi = 0
    else:
        phi = math.atan2(ygsm1, xgsm1)

    byfact = byimf / 5.0
    bzfact = bzimf / 5.0

    pdyn_0 = 2.0
    pfact = (pdyn / pdyn_0) ** XAPPA - 1.0

    cps = math.cos(psi)
    sps = math.sin(psi)
    cp = math.cos(phi)
    sp = math.sin(phi)
    th1 = 0.0
    th2 = PI

    while abs(th1 - th2) > ERR:
        th = 0.5 * (th1 + th2)

        xgsm1 = r * math.sin(th) * cp
        ygsm1 = r * math.sin(th) * sp
        zgsm1 = r * math.cos(th)
        xsm = xgsm1 * cps - zgsm1 * sps
        ysm = ygsm1
        zsm = xgsm1 * sps + zgsm1 * cps

        rho = math.sqrt(xsm**2 + ysm**2)

        if abs(rho) < 1e-6:
            cosphi = 1.0
            sinphi = 0.0
        else:
            cosphi = xsm / rho
            sinphi = ysm / rho

        alpha = ALPHA0 + DALPHA1 * cosphi + DALPHA2 * pfact + DALPHA3 * bzfact
        beta = BETA0 + BETA1 * bzfact

        rh = (
            RH0
            + RH1 * pfact
            + RH2 * bzfact
            + (RH3 + RH4 * pfact + RH5 * bzfact) * cosphi
        )
        t = T0 + T1 * pfact
        a0 = A00 + A01 * pfact + A02 * bzfact
        a1 = A10 + A11 * pfact + A12 * bzfact

        f = 1.0 - (1.0 + (rho / rh) ** alpha) ** (1.0 / alpha)
        g = a0 + a1 * cosphi
        f1 = t * byfact * (rho / 10.0) ** beta * sinphi

        zsm_sheet = rh * math.tan(psi) * f * g + f1

        ff = zsm - zsm_sheet

        if ff > 0:
            th1 = th
        else:
            th2 = th

    theta_s_gsm = 0.5 * (th1 + th2)

    xgsm_s = r * math.sin(theta_s_gsm) * cp
    ygsm_s = r * math.sin(theta_s_gsm) * sp
    zgsm_s = r * math.cos(theta_s_gsm)

    return xgsm_s, ygsm_s, zgsm_s

def tag14_ns_model(time, gsm_pos, pdyn=0.0, byimf=0.0, bzimf=0.0, sc2NS=False):
    """ Wrapper to wrangle parameters for tag_equat_sheet()

    Parameters
    ----------
    time : array_like of str or float
        Time in either string or double format.
        If float, represents seconds since 1970.
        If string, should be in the format: "YYYY-MM-DD/hh:mm:ss".
    gsm_pos : array_like
        Position vector in GSM coordinates in re (pos[*,3]).
    pdyn : float, optional
        Dynamic pressure in nanoPascal
    byimf : float, optional
        Y component of the interplanetary magnetic field, in nT, averaged over previous 30 minutes
    bzimf : float, optional
        Z component of the interplanetary magnetic field, in nT, averaged over previous 30 minutes
    sc2NS : bool, optional
        If set, returns Z displacement from the spacecraft to the neutral sheet.
        Value is positive if the NS is northward of the SC location, and negative if below. Default is False.

    Returns
    -------
    array_like of float
        Returns Z displacement of the neutral sheet above or below the XY plane in Re (zgsm of the NS).
        Value is positive if NS is above z=0 gsm plane, negative if below.
    """
    dz2NS = np.zeros(len(time))

    for i in range(len(time)):
        done = 0
        xgsm = gsm_pos[i, 0]
        ygsm = gsm_pos[i, 1]
        zgsm = gsm_pos[i, 2]

        # get tilt angle of magnetic pole
        tilt = geopack_recalc(time_double(time[i]))
        ns_x, ns_y, ns_z = tag14_equat_sheet(xgsm, ygsm, zgsm, tilt, pdyn, byimf, bzimf)
        dz2NS[i] = ns_z

    if sc2NS:
        scz = gsm_pos[:, 2]
        return dz2NS - scz
    else:
        return dz2NS

def neutral_sheet(
    time, pos, kp=None, model="themis", mlt=None, in_coord="gsm", pdyn=2.0, byimf=0.0, bzimf=0.0, sc2NS=False,
):
    """
    Calculate the GSM-z coordinate of the neutral sheet for a given time and position.

    Each model may require a different subset of the parameters defined here.

    Parameters
    ----------
    time : datetime
        Time of interest.

    pos : array_like
        Position of interest.

    kp : array_like, optional
        Kp index.
        Default: None

    model : str, optional
        Neutral sheet model to use. Valid options are 'sm', 'themis', 'aen', 'den', 'fairfield', 'den_fairfield', 'lopez', 'tag14'
        Default: 'themis'

    mlt : array_like, optional
        Magnetic local time.
        Default: No

    in_coord : str, optional
        Coordinate system of the input position.
        Default: 'gsm'

    pdyn : float, optional
        Solar wind dynamic pressure in nanoPascals

    byimf : float, optional
        Y component of the interplanetary magnetic field, in nT, averaged over previous 30 minutes

    bzimf : float, optional
        Z component of the interplanetary magnetic field, in nT, averaged over previous 30 minutes

    sc2NS : Bool, optional
        Flag to return spacecraft to neutral sheet distance along z-axis, rather than z-coordinate of neutral sheet
        Default: False


    Returns
    -------
    distance2NS : array_like
        Distance from the GSM XY plane (or input position, if sc2NS is True) to the neutral sheet


    Examples
    --------
        # This is a simple example, not a realistic demonstration of the model.
        import numpy as np
        from pyspedas.tplot_tools import time_double

        time = np.array(
            time_double(["2002-02-03T04:05:00", "2002-02-03T04:00:04", "2002-02-03T04:00:08"]),
        )

        # Define a position
        gsm_pos = np.array(
            [
                [6.0, -2.0, 0.5],
                [7.5, 1.5, -0.5],
                [5.5, 0.0, 1.0],
            ]
        )

        # Define a Kp index
        kp = np.array([1, 2, 3])

        # Call the neutral_sheet function
        distance2NS = neutral_sheet(
            time=time, pos=gsm_pos, kp=kp, model="themis", in_coord="gsm", sc2NS=False
        )

        print(distance2NS)
    """

    time = np.array(time)

    # validate and initialize parameters if not set
    if model is None:
        model = "themis"
    else:
        model = model.lower()
    models = ["sm", "themis", "aen", "den", "fairfield", "den_fairfield", "lopez", "tag14"]
    if model not in models:
        logging.error(
            "An invalid neutral sheet model name was used. Valid entries include: "
        )
        logging.error(models)
        return

    # check input coordinate system, convert to gsm if needed
    if in_coord is None:
        in_coord = "gsm"
    else:
        in_coord = in_coord.lower()
    if in_coord == "gsm":
        gsm_pos = pos
    else:
        gsm_pos = cotrans(time_in=time, data_in=pos, coord_in=in_coord, coord_out="gsm")

    # call the appropriate neutral sheet model
    if model == "sm":
        return sm_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == "aen":
        return aen_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == "den":
        return den_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == "fairfield":
        return fairfield_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == "den_fairfield":
        return den_fairfield_ns_model(time, gsm_pos, sc2NS=sc2NS)
    elif model == "lopez":
        return lopez_ns_model(time, gsm_pos, kp=kp, mlt=mlt, sc2NS=sc2NS)
    elif model == "tag14":
        return tag14_ns_model(time, gsm_pos, pdyn=pdyn, byimf=byimf, bzimf=bzimf, sc2NS=sc2NS)
    else:  # model == "themis", the default
        return themis_ns_model(time, gsm_pos, sc2NS=sc2NS)



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyspedas
import copy

def reduce_spec_dataset(tplot_dataset=None, name=None):
    # This function will reduce the data in a 3+ dimensional DataSet object into something that can be plotted with a
    # spectrogram, either by taking slices of the data or by summing the dimensions into this one.
    if tplot_dataset is not None:
        da = copy.deepcopy(tplot_dataset)
    elif name is not None:
        da = copy.deepcopy(pyspedas.tplot_tools.data_quants[name])
    else:
        return

    if da.attrs['plot_options']['extras'].get('spec_dim_to_plot', None) is not None:
        coordinate_to_plot = da.attrs['plot_options']['extras']['spec_dim_to_plot']
    else:
        # If not found, default to v2 (to match behavior in options.py when setting "spec" option)
        coordinate_to_plot = "v2"

    dim_to_plot = coordinate_to_plot + '_dim'

    for d in da.dims:
        if d == dim_to_plot:
            pass
        elif d == 'time':
            pass
        else:
            if 'spec_slices_to_use' in da.attrs['plot_options']['extras']:
                for key, value in da.attrs['plot_options']['extras']['spec_slices_to_use'].items():
                    dim = key+"_dim"
                    if dim == d:
                        da=da.isel({dim:value})
                        break
                else:
                    da = da.sum(dim=d, skipna=True, keep_attrs=True)
            else:
                da = da.sum(dim=d, skipna=True, keep_attrs=True)
    return da




import unittest
from unittest.mock import patch
import pyspedas
from pyspedas.projects.secs.makeplots import make_plots
from pyspedas.tplot_tools import del_data
from pyspedas.projects.secs.config import CONFIG


class SECSTestCases(unittest.TestCase):
    @patch("matplotlib.pyplot.show")
    def test_load_secs(self, mock_show):
        del_data()
        trange = ["2017-03-27", "2017-03-28"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            resolution=10,
            dtype="SECS",
            no_download=False,
            downloadonly=True,
            out_type="df",
            spdf=True,
            force_download=True,
        )
        # Returns a Pandas dataframe, not tplot variables
        self.assertTrue(len(d) > 0)
        make_plots(
            dtype="SECS",
            dtime="2017-03-27/06:00:00",
            vplot_sized=True,
            contour_den=201,
            s_loc=False,
            quiver_scale=30,
        )

    @patch("matplotlib.pyplot.show")
    def test_load_eics(self, mock_show):
        del_data()
        trange = ["2017-03-27", "2017-03-28"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            resolution=10,
            dtype="EICS",
            no_download=False,
            downloadonly=False,
            out_type="df",
            spdf=True,
            force_download=True,
        )
        # Returns a Pandas dataframe, not tplot variables
        self.assertTrue(len(d) > 0)
        make_plots(
            dtype="EICS",
            dtime="2017-03-27/06:00:00",
            vplot_sized=True,
            contour_den=201,
            s_loc=False,
            quiver_scale=30,
        )

    def test_load_secs_dc(self):
        trange = ["2017-03-27", "2017-03-28"]
        del_data()
        d = pyspedas.projects.secs.data(
            trange=trange,
            resolution=10,
            dtype="SECS",
            no_download=False,
            downloadonly=False,
            out_type="dc",
        )
        # Returns a Pandas dataframe, not tplot variables
        self.assertTrue(len(d) > 0)

    def test_load_eics_dc(self):
        del_data()
        trange = ["2017-03-27", "2017-03-28"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            resolution=10,
            dtype="EICS",
            no_download=False,
            downloadonly=False,
            out_type="dc",
        )
        # Returns a Pandas dataframe, not tplot variables
        self.assertTrue(len(d) > 0)

    def test_load_secs_np(self):
        del_data()
        trange = ["2017-03-27", "2017-03-28"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            resolution=10,
            dtype="SECS",
            no_download=False,
            downloadonly=False,
            out_type="np",
        )
        # Returns a Pandas dataframe, not tplot variables
        self.assertTrue(len(d) > 0)

    def test_load_eics_np(self):
        del_data()
        trange = ["2017-03-27", "2017-03-28"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            resolution=10,
            dtype="EICS",
            no_download=False,
            downloadonly=False,
            out_type="np",
        )
        # Returns a Pandas dataframe, not tplot variables
        self.assertTrue(len(d) > 0)

    def test_load_gz(self):
        # A few dates in 2007 have .gz files (instead of .zip files)
        # This only happens for the UCLA data, not the SPDF data
        del_data()
        dtype = "EICS"  # 'EICS or SECS'
        trange = ["2007-02-09/02:15:35", "2007-02-09/02:15:35"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            dtype=dtype,
            downloadonly=True,
        )
        self.assertTrue(len(d) > 0)

    def test_none_cases(self):
        # No data type provided
        del_data()
        trange = ["2017-03-27", "2017-03-28"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            dtype=None,
        )
        self.assertTrue(d is None)

        # Wrong data type provided
        del_data()
        trange = ["2017-03-27", "2017-03-28"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            dtype="aaa",
        )
        self.assertTrue(d is None)

        # Invalid time range provided
        del_data()
        trange = ["2018-03-27", "2017-03-28"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            dtype="dc",
        )
        self.assertTrue(d is None)

        # Invalid out type provided
        trange = ["2017-03-27", "2017-03-28"]
        d = pyspedas.projects.secs.data(
            trange=trange,
            dtype="dc",
            out_type="aaa",
        )
        self.assertTrue(d is None)


if __name__ == "__main__":
    unittest.main()



from pyspedas.projects.elfin.load import load

def fgm_load(trange=['2022-08-19', '2022-08-19'],
        probe='a',
        datatype='survey',
        level='l1',
        suffix='',
        get_support_data=False,
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        force_download=False):
    """
    This function loads data from the ELFIN Fluxgate Magnetometer (FGM)

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2022-08-19', '2022-08-19']

        probe: str
            Spacecraft identifier ('a' or 'b')
            Default: 'a'

        datatype: str
            Data type; Valid options::

                'fast', 'survey' for L1 data. Only 'survey' data is available.

            Default: 'survey'

        level: str
            Data level; options: 'l1'
            Default: l1

        suffix: str
            The tplot variable names will be given this suffix.
            Default: no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: all variables are loaded in.

        varnames: list of str
            List of variable names to load
            Default: all data variables are loaded

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    ----------
        List of tplot variables created.
            ela_fgs
            ela_fgs_fsp_res_dmxl
            ela_fgs_fsp_res_gei
            ela_fgs_fsp_res_ndw
            ela_att_gei_fsp_interp
            ela_fgs_fsp_res_obw

    Example
    ----------
        import pyspedas
        from pyspedas.tplot_tools import tplot
        fgm_vars = pyspedas.projects.elfin.fgm(probe='a', trange=['2022-08-19', '2022-08-19'])
        tplot(['ela_fgs_fsp_res_ndw', 'ela_fgs_fsp_res_obw', 'ela_att_gei_fsp_interp'])

    """

    tvars = load(instrument='fgm', probe=probe, trange=trange, level=level,
                 datatype=datatype, suffix=suffix, get_support_data=get_support_data,
                 varformat=varformat, varnames=varnames, downloadonly=downloadonly,
                 notplot=notplot, time_clip=time_clip, no_update=no_update, force_download=force_download)

    if tvars is None or notplot or downloadonly:
        return tvars

    return fgm_postprocessing(tvars)


def fgm_postprocessing(variables):
    """
    Placeholder for FGM post-processing
    """
    return variables




from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence, Tuple

import matplotlib.dates
import numpy as np
import pyspedas
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pyspedas.tplot_tools import get_data
from pyspedas.tplot_tools import get_var_label_ticks
from matplotlib import pyplot as plt

def var_label_panel(
        variables: Sequence[Any],
        var_label_list: Sequence[Any],
        axs: Sequence[Axes],
        font_size: float
) -> None:

    # Postprocess var label panel
    last_data_idx = len(variables) - 1
    last_data_ax = axs[last_data_idx]

    # List[float] (Number of days since epoch)
    xaxis_ticks = last_data_ax.get_xticks().tolist()

    # Get formatted time strings from Matplotlib
    # The format is not identical to IDL version
    # but is suitable for showing in Matplotlib
    # They are the same as xticklabels,
    # but xticklabels is initialized only after figure is drawn
    # So you need to directly use formatter to get the strings
    locator = last_data_ax.xaxis.get_major_locator()
    formatter = matplotlib.dates.ConciseDateFormatter(locator)
    # Ex. 12:34
    time_suffixes = formatter.format_ticks(xaxis_ticks)
    # Ex. 2017-Apr-01
    time_prefix = formatter.get_offset()

    # List[np.datetime64]
    # num2date returns datetime with timezone info (UTC), so make it None
    # to suppress deprecate warning
    xaxis_ticks_dt = [
        np.datetime64(
            matplotlib.dates.num2date(tick_val).replace(tzinfo=None).isoformat(),
            'ns'
        )
        for tick_val in xaxis_ticks
    ]
    xaxis_ticks_dt_str = [
        str(tick_val.astype(str)) for tick_val in xaxis_ticks_dt
    ]

    # Var label panel
    var_label_axis = axs[last_data_idx + 1]
    var_label_axis.spines["top"].set_visible(False)
    var_label_axis.spines["right"].set_visible(False)
    var_label_axis.spines["bottom"].set_visible(False)
    var_label_axis.spines["left"].set_visible(False)
    var_label_axis.tick_params(axis="x", which="both", length=0, labelbottom=False)
    var_label_axis.tick_params(
        axis="y", which="both", length=0, pad=font_size * 3, labelsize=font_size
    )
    var_label_axis.set_ylim(0, len(var_label_list) + 2)

    ys = []
    y_labels = []
    for i in range(len(var_label_list) + 2):
        y = len(var_label_list) + 2 - 0.5 - i
        # print('i, len(var_label_list)+1', i, len(var_label_list)+1)
        # Time prefix (Ex. 2017-Apr-01) is plotted at y = 0.5 (bottom)
        if i == len(var_label_list) + 1:
            y_label = ""
            _, xmax = var_label_axis.get_xlim()
            var_label_axis.text(
                xmax, y, time_prefix, fontsize=font_size, ha="right", va="center"
            )
        else:
            xmin, xmax = var_label_axis.get_xlim()
            # Time suffix (Ex. 12:34) is plotted at y = 1.5 (bottom)
            if i == len(var_label_list):
                y_label = ""
                xaxis_labels = time_suffixes
            # Other var labels are plotted above time suffix
            else:
                label = var_label_list[i]
                # print('label ', label)
                label_data = get_data(label, xarray=True, dt=True)
                y_label = label_data.attrs["plot_options"]["yaxis_opt"]["axis_label"]  # type: ignore

                # Switch the type of the time array according to the version of numpy
                if np.__version__ >= "2.0.0":
                    xaxis_labels = get_var_label_ticks(
                        label_data, xaxis_ticks_dt_str
                    )
                else:
                    xaxis_labels = get_var_label_ticks(
                        label_data, xaxis_ticks_dt
                    )
                    # type ignore

            for xaxis_tick, xaxis_label in zip(xaxis_ticks, xaxis_labels):  # type: ignore
                # Sometimes ticks produced by locator can be outside xlim, so let exclude them
                if xmin <= xaxis_tick <= xmax:
                    var_label_axis.text(
                        xaxis_tick,
                        y,
                        xaxis_label,
                        fontsize=font_size,
                        ha="center",
                        va="center",
                    )
            ys.append(y)
            y_labels.append(y_label)
    var_label_axis.set_yticks(ys, y_labels, ha="right")




import logging
import numpy as np
from pyspedas.cotrans_tools.gsm2lmn import gsm2lmn
from .rotmat_set_coords import rotmat_set_coords
from pyspedas.tplot_tools import get_data, store_data, set_coords, tnames, options, time_clip
from pyspedas import tinterpol
from pyspedas.projects.omni.omni_solarwind_load import omni_solarwind_load


def lmn_matrix_make(
    pos_var_name,
    mag_var_name,
    trange=None,
    hro2=False,
    newname=None,
):
    """
    Generate an LMN (boundary-normal) coordinate transformation matrix from the given magnetic field B and position data,
    and store it in a tplot variable using the GSM to LMN transformation.

    Parameters
    ----------
    pos_var_name : str
        Name of tplot variable containing the spacecraft position data in GSM coordinates.
    mag_var_name : str
        Name of tplot variable containing the magnetic field data in GSM coordinates.
    trange : list, optional
        Time range to use for interpolation.
        If None, the time range will be set from the mag_var_name tplot variable.
    hro2 : bool, optional
        Flag indicating whether to use the HRO2 level of OMNI data.
        If False, the HRO level will be used.
        Default is False.
     newname : str, optional
        Name of the output tplot variable. Defaults to mag_var_name + "_lmn_mat".

    Returns
    -------
    str or None
        The name of the output tplot variable containing the LMN matrix,
        or None if an error occurs.
    """

    # Check if the variables exist in tplot
    if not tnames(pos_var_name):
        logging.error(f"lmn_matrix_make requires pos_var_name to be set")
        return None
    if not tnames(mag_var_name):
        logging.error(f"lmn_matrix_make requires mag_var_name to be set")
        return None
    
    # Set OMNI level (hro or hro2)
    if hro2:
        level = "hro2"
    else:
        level = "hro"

    # Set the name of the output variable
    if newname is None:
        newname = mag_var_name + "_lmn_mat"

    # Set trange if not provided
    if trange is None or len(trange) != 2:
        mag_var_name_data = get_data(mag_var_name)
        if mag_var_name_data is None:
            logging.error(f"Error reading tplot variable: {mag_var_name}")
            return None
        trange = [mag_var_name_data.times[0], mag_var_name_data.times[-1]]

    # Get solarwind data
    swind = omni_solarwind_load(trange=trange, level=level, min5=False)

    if swind is not None and len(swind) == 2:
        # Interpolate solar wind data to match themagnetic field data
        bz_name = swind[0] + "_interpol"
        p_name = swind[1] + "_interpol"
        tinterpol(swind[0], mag_var_name, newname=bz_name, extrapolate=True)
        tinterpol(swind[1], mag_var_name, newname=p_name, extrapolate=True)

        bz_data = get_data(bz_name)
        p_data = get_data(p_name)
        if bz_data is None or p_data is None:
            logging.error(f"Error reading tplot variables: {bz_name}, {p_name}")
            return None
        timesw = bz_data.times
        dpnon = p_data.y
        bnon = bz_data.y
        swdata = np.array([timesw, dpnon, bnon]).T
    else:
        logging.warning("OMNI Solar Wind data could not be loaded. Exiting.")
        return None

    # Get position and magnetic field data
    pos_var_name_interpol = pos_var_name + "_interpol"
    tinterpol(pos_var_name, mag_var_name, newname=pos_var_name_interpol, extrapolate=True)
    pos_data = get_data(pos_var_name_interpol)
    mag_data = get_data(mag_var_name)
    if pos_data is None or mag_data is None:
        logging.error(f"Error reading tplot variables: {pos_var_name_interpol}, {mag_var_name}")
        return None
    times = mag_data.times
    rxyz = pos_data.y
    bxyz = mag_data.y

    # Perform the GSM to LMN transformation
    try:
        blmn = gsm2lmn(times, rxyz, bxyz, swdata=swdata)
    except Exception as e:
        logging.error(f"Error during GSM to LMN transformation: {e}")
        return None

    # Store the LMN transformation matrix in a new tplot variable
    store_data(newname, data={"x": times, "y": blmn})
    rotmat_set_coords(newname, "GSM", "LMN")
    options(newname, "ysubtitle", "[LMN]")
    options(newname, "legend_names", ["BL", "BM", "BN"])

    time_clip(newname, trange[0], trange[1], overwrite=True)

    return newname



import os
import logging
import pyspedas
from pyspedas.mth5.config import CONFIG

# Make sure that download folder exists
if not os.path.exists(CONFIG['local_data_dir']):
    os.makedirs(CONFIG['local_data_dir'])

# Remove imported
del os, CONFIG

# Determine if MTH5 is in the scope
try:
    from mth5 import config as mth5_logger_config

except (ImportError, ModuleNotFoundError):
    logging.error(f'MTH5 must be installed to use module {__name__}.')
    logging.error("Please install it using: pip install mth5")
    raise

# # Synchronize mth5 logging output level with pyspedas logging output level
try:
    import loguru, pyspedas
    from mth5 import config as mth5_logger_config
    from pyspedas import logging_level
    from pyspedas.mth5.load_fdsn import _disable_loguru_warnings

    # This is how to disable all together
    # import warnings
    # warnings.filterwarnings('ignore')

    # TODO: terminate this code if handler logging_level is the same as in loguru

    mth5_logger_config['handlers'][0]['level'] = logging_level
    mth5_logger_config['handlers'][0]["filter"] = _disable_loguru_warnings
    mth5_logger_config['extra']['no_warning'] = False
    if loguru.logger._core.handlers:
        handler_id = next(iter(loguru.logger._core.handlers.keys()))
        loguru.logger.remove(handler_id)
        loguru.logger.configure(**mth5_logger_config)
        logging.debug('loguru synchronized with pyspedas logging')
except:
    logging.error('loguru synchronization error')
    pass


import numpy as np 
from pyspedas.tplot_tools import get_data, store_data
from pyspedas.tplot_tools import tnames
from pyspedas.tplot_tools import data_exists


def mms_fgm_remove_flags(probe, data_rate, level, instrument, suffix=''):
    """
    This function removes data flagged by the FGM 'flag' variable (flags > 0), 
    in order to only show science quality data by default.
    
    Parameters
    -----------
        probe: str or list of str
            probe or list of probes, valid values for MMS probes are ['1','2','3','4']. 

        data_rate: str or list of str
            instrument data rates for FGM include 'brst' 'fast' 'slow' 'srvy'. The
            default is 'srvy'.

        level: str
            indicates level of data processing. the default if no level is specified is 'l2'

        instrument: str
            instrument; probably 'fgm'
            
        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

    """
    if not isinstance(probe, list): probe = [probe]
    if not isinstance(data_rate, list): data_rate = [data_rate]
    if not isinstance(level, list): level = [level]

    tplot_vars = set(tnames())

    for this_probe in probe:
        for this_dr in data_rate:
            for this_lvl in level:
                if level == 'ql':
                    flag_var = 'mms'+str(this_probe)+'_'+instrument+'_'+this_dr+'_'+this_lvl+'_flag'+suffix
                else:
                    flag_var = 'mms'+str(this_probe)+'_'+instrument+'_flag_'+this_dr+'_'+this_lvl+suffix

                if not data_exists(flag_var):
                    continue
                    
                flagged = get_data(flag_var, dt=True)

                if flagged is None:
                    continue

                times, flags = flagged
                flagged_data = np.where(flags != 0.0)[0]

                for var_specifier in ['_b_gse_', '_b_gsm_', '_b_dmpa_', '_b_bcs_']:
                    var_name = 'mms'+str(this_probe)+'_'+instrument+var_specifier+this_dr+'_'+this_lvl+suffix
                    if var_name in tplot_vars:
                        times, var_data = get_data(var_name, dt=True)
                        metadata = get_data(var_name, metadata=True)
                        var_data[flagged_data] = np.nan
                        store_data(var_name, data={'x': times, 'y': var_data}, attr_dict=metadata)



from pyspedas.utilities.datasets import find_datasets

# This routine was originally in polar/__init__.py.  If you need to see the history of this routine before
# it was moved to its own file, please check the history for __init__.py.

def datasets(instrument=None, label=True):
    return find_datasets(mission='Polar', instrument=instrument, label=label)



"""
Loads geomagnetic index data into tplot variables.

Data is downloaded from the NOAA or GFZ (Helmholtz Centre Potsdam) ftp sites.

Data is stored in tplot variables with the following names:

["Kp", "ap", "Sol_Rot_Num", "Sol_Rot_Day", "Kp_Sum", "ap_Mean", "Cp", "C9", "Sunspot_Number", "F10.7", "Flux_Qualifier"]
"""

import logging, os
import numpy as np
from pyspedas.tplot_tools import (
    time_string,
    time_double,
    store_data,
    options,
    time_clip as tclip,
)
from pyspedas.projects.noaa.config import CONFIG
from pyspedas.utilities.download_ftp import download_ftp
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download


def kp_return_fraction(value):
    """
    Calculate the fractional part of the Kp index.

    Parameters
    ----------
    value : int
        The Kp value in integer format.

    Returns
    -------
    list of float
        The fractional part of the Kp index.
    """
    value = np.array(value, dtype=np.float64)
    kp_lhs = value // 10
    kp_rhs_times_3 = value % 10
    kp_rhs = kp_rhs_times_3 // 3.0
    return kp_lhs + kp_rhs / 3.0


def convert_to_float_or_int(a, outtype="int"):
    """
    Converts a list of strings to either float or int.

    Parameters
    ----------
    a : list of str
        The list of values to be converted.
    outtype : str
        The desired output type, "float" or "int". Default is "int".

    Returns
    -------
    list of int or float:
        The converted list of values.
    """
    ans = []
    for v in a:
        try:
            if outtype == "float":
                ans.append(float(v))
            else:  # outtype == "int"
                ans.append(int(v))
        except ValueError:
            ans.append(0)
    return ans


def load_kp_to_tplot(
    trange=[0, 0],
    files=[],
    datatype=[],
    prefix="",
    suffix="",
):
    """
    Load Kp index data into tplot variables.

    These are text files that have to be parsed to extract the data and store it in tplot variables.

    Parameters
    ----------
    trange : list of str
        Time range to load data for. Should be a list of two strings.
    datatype : list of str, optional
        Type of index to load. Default is an empty list, which loads all available data.

    Returns
    -------
    list of str
        List of tplot variables that contain the data.
    """
    vars = []

    if files is None or len(files) == 0:
        logging.error("No files specified")
        return vars
    elif not isinstance(files, list):
        files = [files]

    if datatype is None or datatype == [] or len(datatype) == 0:
        datatype = [
            "Kp",
            "ap",
            "Sol_Rot_Num",
            "Sol_Rot_Day",
            "Kp_Sum",
            "ap_Mean",
            "Cp",
            "C9",
            "Sunspot_Number",
            "F10.7",
            "Flux_Qualifier",
        ]
    elif not isinstance(datatype, list):
        datatype = [datatype]
    datatype = [d.lower() for d in datatype]

    # Each line in the files contains data for one day.
    # The first 3 quantities are measured every 3 hours (8 measurements per day).
    kpdata = []
    apdata = []
    kptimes = []

    daytimes = []
    srndata = []
    srddata = []
    kpsdata = []
    apmdata = []
    cpdata = []
    c9data = []
    ssndata = []
    srfdata = []
    fqdata = []

    for kpfile in files:
        # Example line contained in these files:
        # 1701012502 73337272323302017210 18 22 12  9  9 15  7  6 120.73---070.10
        # Line length is 73 for NOAA files and 63 for WDC files (both counts include \n).

        with open(kpfile, "r") as f:
            for line in f:
                if len(line) < 63 or line.startswith("#") or line.startswith(" "):
                    # Skip lines that are less than 63 characters long (62 data and \n)
                    continue

                # Get datetimes (0:6 characters)
                year = line[0:2]
                if "00" <= year <= "69":
                    year = "20" + year
                else:
                    year = "19" + year
                month = line[2:4]
                day = line[4:6]
                ymd = year + "-" + month + "-" + day
                daytimes.append(time_double(ymd))
                for k in range(8):
                    dd = time_double(ymd + " " + "{:02d}".format(k * 3) + ":00:00")
                    kptimes.append(dd)

                # Get data (6:71 characters)
                srndata.append(line[6:10])
                srddata.append(line[10:12])
                for k in range(8):
                    kpdata.append(line[12 + 2 * k : 14 + 2 * k])
                kpsdata.append(line[28:31])
                for k in range(8):
                    apdata.append(line[31 + 3 * k : 34 + 3 * k])
                apmdata.append(line[55:58])
                cpdata.append(line[58:61])
                c9data.append(line[61:62])
                if len(line) >= 72:
                    ssndata.append(line[62:65])
                    srfdata.append(line[65:70])
                    fqdata.append(line[70:71])

    # At this point, list are strings. Convert to numbers.
    srndata = convert_to_float_or_int(srndata)
    srddata = convert_to_float_or_int(srddata)
    kpdata = convert_to_float_or_int(kpdata)
    kpsdata = convert_to_float_or_int(kpsdata)
    apdata = convert_to_float_or_int(apdata)
    apmdata = convert_to_float_or_int(apmdata)
    cpdata = convert_to_float_or_int(cpdata, "float")
    c9data = convert_to_float_or_int(c9data)
    ssndata = convert_to_float_or_int(ssndata)
    srfdata = convert_to_float_or_int(srfdata, "float")
    fqdata = convert_to_float_or_int(fqdata)

    # Check for empty data set. If empty, return.
    if len(kptimes) == 0:
        logging.error("No data found.")
        return vars
    if (
        len(trange) == 2
        and trange[0] != 0
        and time_double(trange[1]) > time_double(trange[0])
        and (
            time_double(trange[1]) < kptimes[0] or time_double(trange[0]) > kptimes[-1]
        )
    ):
        logging.error("No data found in time range.")
        return vars

    # Store tplot variables
    # kp (3-hour Kp)
    if "kp" in datatype:
        vn = prefix + "Kp" + suffix
        store_data(vn, data={"x": kptimes, "y": kp_return_fraction(kpdata)})
        options(vn, "ytitle", "Kp")
        vars.append(vn)
    # ap (3-hour ap)
    if "ap" in datatype:
        vn = prefix + "ap" + suffix
        store_data(vn, data={"x": kptimes, "y": apdata})
        options(vn, "ytitle", "ap")
        vars.append(vn)
    # sol_rot_num (solar rotation number)
    if "sol_rot_num" in datatype:
        vn = prefix + "Sol_Rot_Num" + suffix
        store_data(vn, data={"x": daytimes, "y": srndata})
        options(vn, "ytitle", "Sol_Rot_Num")
        vars.append(vn)
    # sol_rot_day (day of solar rotation)
    if "sol_rot_day" in datatype:
        vn = prefix + "Sol_Rot_Day" + suffix
        store_data(vn, data={"x": daytimes, "y": srddata})
        options(vn, "ytitle", "Sol_Rot_Day")
        vars.append(vn)
    # kp_sum (sum of Kp)
    if "kp_sum" in datatype:
        vn = prefix + "Kp_Sum" + suffix
        store_data(vn, data={"x": daytimes, "y": kpsdata})
        options(vn, "ytitle", "Kp_Sum")
        vars.append(vn)
    # ap_mean (mean of Ap)
    if "ap_mean" in datatype:
        vn = prefix + "ap_Mean" + suffix
        store_data(vn, data={"x": daytimes, "y": apmdata})
        options(vn, "ytitle", "ap_Mean")
        vars.append(vn)
    # cp (planetary C index)
    if "cp" in datatype:
        vn = prefix + "Cp" + suffix
        store_data(vn, data={"x": daytimes, "y": cpdata})
        options(vn, "ytitle", "Cp")
        vars.append(vn)
    # c9 (9-point C index)
    if "c9" in datatype:
        vn = prefix + "C9" + suffix
        store_data(vn, data={"x": daytimes, "y": c9data})
        options(vn, "ytitle", "C9")
        vars.append(vn)

    # The following do not exist in the GFZ data
    # sunspot_number (sunspot number)
    lentimes = len(daytimes)
    if "sunspot_number" in datatype and len(ssndata) > 0 and len(ssndata) == lentimes:
        vn = prefix + "Sunspot_Number" + suffix
        store_data(vn, data={"x": daytimes, "y": ssndata})
        options(vn, "ytitle", "Sunspot_Number")
        vars.append(vn)
    # f10.7 (10.7 cm solar radio flux)
    if "f10.7" in datatype and len(srfdata) > 0 and len(srfdata) == lentimes:
        vn = prefix + "F10.7" + suffix
        store_data(vn, data={"x": daytimes, "y": srfdata})
        options(vn, "ytitle", "F10.7")
        vars.append(vn)
    # flux_qualifier (10.7 cm solar radio flux qualifier)
    if "flux_qualifier" in datatype and len(fqdata) > 0 and len(fqdata) == lentimes:
        vn = prefix + "Flux_Qualifier" + suffix
        store_data(vn, data={"x": daytimes, "y": fqdata})
        options(vn, "ytitle", "Flux_Qualifier")
        vars.append(vn)

    return vars


def noaa_load_kp(
    trange=["2017-03-01/00:00:00", "2017-03-31/23:59:59"],
    local_kp_dir=None,
    datatype=[],
    gfz=False,
    prefix="",
    suffix="",
    time_clip=True,
    force_download=False,
):
    """
    Load geomagnetic index data into appropriate variables.


    Parameters
    ----------
    trange : list of str
        Time range to load data for. Should be a list of two strings.
    local_kp_dir : str
        Directory where data is saved locally.
    datatype : list of str, optional
        Type of index to load. Default is an empty list, which loads all available data. Valid values:
            "Kp",
            "ap",
            "Sol_Rot_Num",
            "Sol_Rot_Day",
            "Kp_Sum",
            "ap_Mean",
            "Cp",
            "C9",
            "Sunspot_Number",
            "F10.7",
            "Flux_Qualifier",
    gfz : bool, optional
        Load data from the HTTPS site of the German Research Centre for Geosciences, instead of NOAA.
        This is the default behavior if the end time is on or after 2018.
        Default is False (by default, it loads data from the HTTPS server of NOAA).
        If this source is used, the Sunspot_Number, F10.7, and Flux_Qualifier datatypes will not be available.
    prefix: str, optional
        If provided, specifies a string to be prepended to tplot variable names.
    suffix: str, optional
        If provided, specifies a string to be appended to tplot variable names.
    time_clip: bool, optional
        If True, data will be time clipped to the exact time range requested.
        Default is True.
    force_download: bool
        Download file even if local version is more recent than server version.
        Default is False.

    Returns
    -------
    list of str
        Returns the names of tplot variables that contain all the data.
        Data is saved in a tplot variables.

    Example
    -------
    >>> from pyspedas import noaa_load_kp
    >>> vars = noaa_load_kp(trange=['2014-03-23/00:00:00', '2014-03-23/23:59:59'])
    >>> print(vars)
    ['Kp', 'ap', 'Sol_Rot_Num', 'Sol_Rot_Day', 'Kp_Sum', 'ap_Mean', 'Cp', 'C9', 'Sunspot_Number', 'F10.7', 'Flux_Qualifier']

    """
    vars = []

    if len(trange) == 2:
        trangestr = time_string(time_double(trange))
        start_year = int(trangestr[0][0:4])
        end_year = int(trangestr[1][0:4])
    else:
        logging.error("Invalid time range")
        return
    if end_year > start_year + 3:  # Limit to 4 years
        end_year = start_year + 3
        trange[1] = str(end_year) + "-12-31/00:00:00"
        logging.warning(
            "Time limit is 4 years, new time range is " + trange[0] + " to " + trange[1]
        )
    elif end_year < start_year:
        end_year = start_year
        trange[1] = str(end_year) + "-12-31/00:00:00"
        logging.warning(
            "End time is before start time, new time range is "
            + trange[0]
            + " to "
            + trange[1]
        )

    # Remote site and directory
    if gfz or end_year >= 2018:
        remote_data_dir = CONFIG['gfz_remote_data_dir']
        pathformat = "Kp_def%Y.wdc"
    else:
        remote_data_dir=CONFIG['noaa_remote_data_dir']
        pathformat = "%Y"
    # Local directory
    if not local_kp_dir:
        local_data_dir = CONFIG["local_data_dir"]
    else:
        local_data_dir = local_kp_dir

    if len(local_data_dir) > 0 and not local_data_dir.endswith(os.sep):
        local_data_dir += os.sep

    logging.info("Loading geomagnetic index data from " + remote_data_dir)
    remote_names = dailynames(file_format=pathformat, trange=trange)
    files = download(
        remote_file=remote_names,
        remote_path=remote_data_dir,
        local_path=local_data_dir,
        force_download=force_download,
    )

    if len(files) == 0:
        logging.error("No files found.")
        return vars

    vars = load_kp_to_tplot(
        trange=trange,
        files=list(set(files)),
        datatype=datatype,
        prefix=prefix,
        suffix=suffix,
    )

    if time_clip:
        tclip(vars, trange[0], trange[1], overwrite=True)

    return vars



import logging
import numpy as np
from scipy.ndimage import shift
from pyspedas.projects.mms.particles.mms_convert_flux_units import mms_convert_flux_units


def moka_mms_clean_data(data_in, units=None, disterr=None):
    """
    This is a translation of Mitsuo Oka's IDL routine: moka_mms_clean_data
    """
    if units is None:
        logging.error('Units must be specified.')
        return

    data = mms_convert_flux_units(data_in, units=units)
    data_psd = mms_convert_flux_units(data_in, units='df_km')

    output = {'charge': data_in['charge'], 'mass': data_in['mass'],
              'data': np.reshape(data_in['data'], [data_in['data'].shape[0]*data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'bins': np.reshape(data_in['bins'], [data_in['data'].shape[0]*data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'theta': np.reshape(data_in['theta'], [data_in['data'].shape[0]*data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'energy': np.reshape(data_in['energy'], [data_in['data'].shape[0], data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'phi': np.reshape(data_in['phi'], [data_in['data'].shape[0]*data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'dtheta': np.reshape(data_in['dtheta'], [data_in['data'].shape[0]*data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'dphi': np.reshape(data_in['dphi'], [data_in['data'].shape[0]*data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'denergy': np.reshape(data_in['denergy'], [data_in['data'].shape[0]*data_in['data'].shape[1]*data_in['data'].shape[2]], order='F')}

    de = output['energy'] - shift(output['energy'], [1, 0])
    output['denergy'] = shift((de+shift(de, [1, 0]))/2.0, -1)
    # just have to make a guess at the edges(bottom edge)
    output['denergy'][0, :] = de[1, :]
    # just have to make a guess at the edges(top edge)
    output['denergy'][-1, :] = de[-1, :]

    dims = data['data'].shape
    imax = dims[0]*dims[1]*dims[2]

    output['energy'] = np.reshape(output['energy'], [output['energy'].shape[0] * output['energy'].shape[1]],
                                  order='F')
    output['denergy'] = np.reshape(output['denergy'], [output['denergy'].shape[0] * output['denergy'].shape[1]],
                                  order='F')

    # Error
    psd = output['data']
    if disterr is None:
        err = np.zeros(imax)
        cnt = np.zeros(imax)
    else:
        data_err = mms_convert_flux_units(disterr, units='df_km')
        err = data_err['data']
        cnt = (psd/err)**2  # actual counts recovered

    # NaN
    dat = output['data']
    bins = output['bins']
    idx = np.argwhere(bins == False)
    if len(idx) > 0:
        dat[idx] = 0.0
        psd[idx] = 0.0
        err[idx] = 0.0
        cnt[idx] = 0.0

    dat = np.nan_to_num(dat, nan=0.0)
    psd = np.nan_to_num(psd, nan=0.0)
    err = np.nan_to_num(err, nan=0.0)
    cnt = np.nan_to_num(cnt, nan=0.0)

    output['data_dat'] = dat
    output['data_psd'] = psd
    output['data_err'] = err
    output['data_cnt'] = cnt
    output['pa'] = np.zeros(imax)

    return output



import pyspedas
from pyspedas.tplot_tools import store_data, tinterp
import copy
import logging


def multiply(tvar1, tvar2, newname=None):
    """
    Multiplies two tplot variables.  Will interpolate if the two are not on the same time cadence.

    Parameters
    ----------
        tvar1 : str
            Name of first tplot variable.
        tvar2 : int/float
            Name of second tplot variable.
        newname : str
            Name of new tplot variable.  If not set, then the data in tvar1 is replaced.

    Returns
    -------
        None

    Examples
    --------
        >>> x1 = [0, 4, 8, 12, 16]
        >>> x2 = [0, 4, 8, 12, 16, 19, 21]
        >>> time1 = [pyspedas.time_float("2020-01-01") + i for i in x1]
        >>> time2 = [pyspedas.time_float("2020-01-01") + i for i in x2]
        >>> pyspedas.store_data("a", data={"x": time1, "y": [1, 2, 3, 4, 5]})
        >>> pyspedas.store_data("c", data={"x": time2, "y": [1, 4, 1, 7, 1, 9, 1]})
        >>> n = pyspedas.multiply("a", "c", newname="a_x_c")
        >>> print('new tplot variable:', n)
        >>> ac = pyspedas.get_data(n)
        >>> print(ac)
    """

    # interpolate tvars
    tv2 = tinterp(tvar1, tvar2)
    # separate and multiply data
    data1 = pyspedas.tplot_tools.data_quants[tvar1].values
    data2 = pyspedas.tplot_tools.data_quants[tv2].values
    data = data1 * data2

    if newname is None:
        pyspedas.tplot_tools.data_quants[tvar1].values = data
        return tvar1

    if "spec_bins" in pyspedas.tplot_tools.data_quants[tvar1].coords:
        store_data(
            newname,
            data={
                "x": pyspedas.tplot_tools.data_quants[tvar1].coords["time"].values,
                "y": data,
                "v": pyspedas.tplot_tools.data_quants[tvar1].coords["spec_bins"].values,
            },
        )
        pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(
            pyspedas.tplot_tools.data_quants[tvar1].attrs
        )
    else:
       store_data(
            newname,
            data={"x": pyspedas.tplot_tools.data_quants[tvar1].coords["time"].values, "y": data},
        )
       pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(
            pyspedas.tplot_tools.data_quants[tvar1].attrs
        )
    return newname



import logging
import pyspedas
import numpy as np


def yclip(
    names,
    ymin,
    ymax,
    flag=None,
    newname=None,
    suffix=None,
    overwrite=None,
):
    """
    Clip data and replace values with flag.

    Parameters
    ----------
    names : str or list of str
        List of tplot names.
        Allowed wildcards are ? for a single character, * from multiple characters.
    ymin : float
        Minimum value to clip.
    ymax : float
        Maximum value to clip.
    flag : float, optional
        Values outside (ymin, ymax) are replaced with this flag.
        Default is float('nan').
    newname : str or list of str, optional
        List of new names for tplot variables.
        If not given, then a suffix is applied.
    suffix : str, optional
        A suffix to apply. Default is '-clip'.
    overwrite : bool, optional
        Replace the existing tplot name.

    Returns
    -------
    None
        This function works in-place and does not return a value, it creates a new tplot variable.

    Notes
    -----
    Similar to tclip.pro in IDL SPEDAS.
    This function clips y-axis data. To clip time-axis, use time_clip.

    Examples
    --------
        import numpy as np
        import pyspedas
        import pyspedas

        # Create some data
        times = np.array(['2002-02-03T04:05:06', '2002-02-03T04:05:07', '2002-02-03T04:05:08'], dtype='datetime64')
        data = np.array([1.0, 2.0, 3.0])

        # Store the data in tplot 'variable1'
        pyspedas.store_data('variable1', data={'x': times, 'y': data})

        # Clip the data between 1.5 and 2.5, replacing values outside this range with NaN
        pyspedas.yclip(names='variable1', ymin=1.5, ymax=2.5)

        # The clipped data is now stored in 'variable1-clip'
        clipped_data = pyspedas.get_data('variable1-clip')
        print(clipped_data)
    """

    old_names = pyspedas.tnames(names)

    if len(old_names) < 1:
        logging.error("yclip: No valid tplot names were provided.")
        return

    if suffix is None:
        suffix = "-clip"

    if flag is None:
        flag = np.nan

    if overwrite is not None:
        n_names = old_names
    elif newname is None:
        n_names = [s + suffix for s in old_names]
    else:
        n_names = newname

    if isinstance(n_names, str):
        n_names = [n_names]

    if len(n_names) != len(old_names):
        n_names = [s + suffix for s in old_names]

    for i, old in enumerate(old_names):
        new = n_names[i]

        _ = pyspedas.clip(old, ymin, ymax, new)

        logging.info("yclip was applied to: " + new)



"""
@Author: Xin Cao, Xiangning Chu, University of Colorado Boulder
@Created in Sep. 2021.

In this config.py file, the following environment variables need to be defined in advance.
@LOCAL_SECS_DATA_DIR: the local path where the SECS/EICS data will be downloaded and stored.
@PLOTS_SECS_DIR: the local path where the vector and contour plots of SECS/EICS data will be saved.
"""

import os

CONFIG = {
    "local_data_dir": "secs_data/",
    "plots_dir": "secs_plots/",
    "remote_data_dir": "http://vmo.igpp.ucla.edu/data1/SECS/",
    "remote_data_dir_spdf": "https://spdf.gsfc.nasa.gov/pub/data/aaa_special-purpose-datasets/spherical-elementary-and-equivalent-ionospheric-currents-weygand/",
}

# override local data directory with environment variables
if os.environ.get("LOCAL_SECS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.environ[
        "LOCAL_SECS_DATA_DIR"
    ]  # self-defined, e.g. '/Users/cao/Documents/Data_N_Res/data_local'
elif os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.sep.join([os.environ["SPEDAS_DATA_DIR"], "secs"])

if os.environ.get("PLOTS_SECS_DIR"):
    CONFIG["plots_dir"] = os.environ[
        "PLOTS_SECS_DIR"
    ]  # self-defined, e.g. '/Users/cao/Documents/Data_N_Res/plots'
elif os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["plots_dir"] = os.sep.join([os.environ["SPEDAS_DATA_DIR"], "secs_plots"])

if os.environ.get("CONDA_PREFIX"):
    env_name = os.environ["CONDA_PREFIX"]
    os.environ["PROJ_LIB"] = env_name + "/share/proj"



import logging
import numpy as np
import matplotlib.pyplot as plt
from .slice2d_getinfo import slice2d_getinfo
from pyspedas.particles.spd_units_string import spd_units_string


def slice1d_plot(the_slice, direction, value, xrange=None, yrange=None):
    """
    Create 1D plot from a 2D particle slice.

    If the 'value' argument is a scalar, this plots a cut through
    the distribution at the nearest point in the specified direction

    If the 'value' argument is a two-element list, this sums over
    the values between the min and max of the list

    Parameters
    ----------
        the_slice: dict
            2D slice returned by slice2d

        direction: str
            Axis to plot - 'x' or 'y'

        value: float or list of float
            If direction is 'x', this is the y-value to create a 1D plot at (and vice versa)
            can also be a range of values, e.g., [-1000, 1000] to sum over the y-values
            from -1000 to +1000
    """

    direction = direction.lower()
    if direction not in ['x', 'y']:
        logging.error('Invalid direction specified. Valid options are: x, y')
        return

    if xrange is None:
        if direction == 'x':
            xrange = [np.nanmin(the_slice['ygrid']), np.nanmax(the_slice['ygrid'])]
        elif direction == 'y':
            xrange = [np.nanmin(the_slice['xgrid']), np.nanmax(the_slice['xgrid'])]

    fig = plt.figure()
    axis = fig.add_subplot(111)

    fig.subplots_adjust(left=0.14, right=0.94, top=0.90, bottom=0.14)

    # get annotations
    annotations = slice2d_getinfo(the_slice)

    xunits = the_slice['xyunits']

    if the_slice['energy']:
        xtitle = 'E (' + xunits + ')'
    else:
        if direction == 'x':
            xtitle = annotations['xtitle']
        else:
            xtitle = annotations['ytitle']

    options = {'linewidth': 0.5,
               'color': 'black'}

    axis.set_xlabel(xtitle)
    axis.set_ylabel(annotations['ztitle'])
    axis.set_yscale('log')

    if isinstance(value, list):
        # sum over a range specified by value
        if direction == 'x':
            values_to_incl = np.argwhere((the_slice['ygrid'] >= value[0]) & (the_slice['ygrid'] <= value[1])).flatten()
            if len(values_to_incl) != 0:
                axis.plot(the_slice['xgrid'], np.nansum(the_slice['data'][:, values_to_incl], axis=1), **options)
        elif direction == 'y':
            values_to_incl = np.argwhere((the_slice['xgrid'] >= value[0]) & (the_slice['xgrid'] <= value[1])).flatten()
            if len(values_to_incl) != 0:
                axis.plot(the_slice['ygrid'], np.nansum(the_slice['data'][values_to_incl, :], axis=0), **options)

        plt.show()
        return

    # single point
    if direction == 'x':
        closest_at_this_value = the_slice['ygrid'].flat[np.abs(the_slice['ygrid'] - value).argmin()]
        idx_at_this_value = np.argwhere(the_slice['ygrid'] == closest_at_this_value).flatten()
        axis.plot(the_slice['ygrid'], the_slice['data'][:, idx_at_this_value].flatten(), **options)
    elif direction == 'y':
        closest_at_this_value = the_slice['xgrid'].flat[np.abs(the_slice['xgrid'] - value).argmin()]
        idx_at_this_value = np.argwhere(the_slice['xgrid'] == closest_at_this_value).flatten()
        axis.plot(the_slice['xgrid'], the_slice['data'][idx_at_this_value, :].flatten(), **options)

    fig.show()
    return

# For backward compatibility with PyHC Summer School 2024 notebooks
plot = slice1d_plot


from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG

def load(trange=['2013-11-5', '2013-11-6'], 
         instrument='reptile',
         datatype='flux', 
         level='l2',
         prefix='',
         suffix='', 
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False,
         force_download=False):
    """
    This function loads data from the CSSWE mission from the Relativistic Electron and Proton
    Telescope integrated little experiment (REPTile). This function is not meant
    to be called directly; instead, see the wrapper:
        pyspedas.projects.csswe.reptile

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2013-11-5', '2013-11-6']

        datatype: str
            Data type; Valid options:
                'counts' for L1 data
                'flux' for L2 data

        level: str
            Data level; options: 'l1', 'l2' (default: l2)

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Returns
    ----------
        List of tplot variables created.

    Example
    ----------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> reptile_csswe_vars = pyspedas.projects.csswe.reptile(trange=['2013-11-5', '2013-11-6'])

    """

    if prefix is None:
        prefix = ''
    if suffix is None:
        suffix = ''

    if instrument == 'reptile':
        pathformat = level+'/'+instrument+'/'+datatype+'/%Y/csswe_'+instrument+'_6sec-'+datatype+'-'+level+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



from pyspedas.projects.themis.load import load
from pyspedas.tplot_tools import data_exists, del_data
from .apply_spinaxis_corrections import apply_spinaxis_corrections
from .spinmodel.spinmodel_postprocess import spinmodel_postprocess


def state(trange=['2007-03-23', '2007-03-24'],
          probe='c',
          level='l1',
          suffix='',
          get_support_data=False,
          varformat=None,
          exclude_format=None,
          varnames=[],
          downloadonly=False,
          notplot=False,
          no_update=False,
          time_clip=False,
          keep_spin=False):
    """
    Load THEMIS state data

    Parameters
    ----------
        trange: list of str
            time range of interest [starttime, endtime] with the format
            ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2007-03-23', '2007-03-24']

        probe: str or list of str
            Spacecraft probe letter(s) ('a', 'b', 'c', 'd' and/or 'e')
            Default: 'c'

        level: str
            Data type; Valid options: 'l1'
            Default: 'l1'

        suffix: str
            The tplot variable names will be given this suffix.
            Default: None

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.
            Default: None (all variables are loaded)

        exclude_format: str
            If specified, CDF variables matching this pattern will not be processed.
            Wildcard character "*" is accepted.
            Default: None

        varnames: list of str
            List of variable names to load. If list is empty or unsoecified, all data variables are loaded
            Default: [] (all variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: false

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: false

        no_update: bool
            If set, only load data from your local cache
            Default: false

        time_clip: bool
            Time clip the variables to exactly the range specified
            in the trange keyword
            Default: false

        keep_spin: bool
            If True, do not delete the spin model tplot variables after the spin models are built.
            Default: False


    Returns
    -------
        List of str
        List of tplot variables created
        Empty list if no data loaded

    Example
    -------
        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> pyspedas.projects.themis.state(trange=['2007-03-23', '2007-03-24'], probe='a', varnames=['tha_pos_gse','tha_vel_gse'])
        >>> tplot['tha_pos_gse', 'tha_vel_gse'])

    """
    # If support data is being loaded, preemptively delete the thx_spinras_correction and thx_spindec_correction
    # variables, to avoid dangling corrections if they don't exist in this time interval.
    if get_support_data:
        for p in probe:
            spinras_corrvar='th'+ p +'_spinras_correction'
            spindec_corrvar='th'+ p +'_spindec_correction'
            if data_exists(spinras_corrvar):
                del_data(spinras_corrvar)
            if data_exists(spindec_corrvar):
                del_data(spindec_corrvar)

    res = load(instrument='state', trange=trange, level=level, probe=probe,
               suffix=suffix, get_support_data=get_support_data,
               varformat=varformat, varnames=varnames, exclude_format = exclude_format,
               downloadonly=downloadonly, notplot=notplot,
               time_clip=time_clip, no_update=no_update)
    if get_support_data:
        for p in probe:
            # Process spin model variables
            spinmodel_postprocess(p,suffix)
            if not keep_spin:
                spinvar_pattern = 'th' + p + '_spin_*'
                del_data(spinvar_pattern)
            # Perform spin axis RA and Dec corrections
            spinras_var = 'th' + p + '_spinras' + suffix
            delta_spinras_var = 'th' + p + '_spinras_correction' + suffix
            corrected_spinras_var = 'th' + p + '_spinras_corrected' + suffix

            spindec_var = 'th' + p + '_spindec' + suffix
            delta_spindec_var = 'th' + p + '_spindec_correction' + suffix
            corrected_spindec_var = 'th' + p + '_spindec_corrected' + suffix

            apply_spinaxis_corrections(spinras=spinras_var, delta_spinras=delta_spinras_var,
                                       corrected_spinras=corrected_spinras_var, spindec=spindec_var,
                                       delta_spindec=delta_spindec_var, corrected_spindec=corrected_spindec_var)
    return res
